from pathlib import Path

from url.url import URL

from bowtie._benchmarks import Benchmark, BenchmarkGroup
from bowtie._core import Dialect


def get_benchmark():
    name = "maxItems"
    benchmark_type = "keyword"
    description = (
        "A benchmark for measuring performance of the "
        "implementation for the maxItems keyword."
    )
    max_array_length = 100000
    varying_parameter = "Array length"

    array_length = 1000
    benchmarks = []
    tests = []
    testing_array_length = 10000

    while testing_array_length <= max_array_length:
        tests.append(
            dict(
                description=f"Testing Array length - {testing_array_length}",
                instance=["random" for _ in range(testing_array_length)],
            ),
        )
        testing_array_length *= 10

    while array_length <= max_array_length:
        benchmarks.append(
            Benchmark.from_dict(
                name=f"Array length - {array_length}",
                description=(
                    f"Validating the `maxItems` keyword over array of length {array_length}."
                ),
                schema={
                    "maxItems": array_length,
                },
                tests=tests,
            ),
        )
        array_length *= 10

    return BenchmarkGroup(
        name=name,
        benchmark_type=benchmark_type,
        description=description,
        dialects_supported=[
            Dialect.from_str("https://json-schema.org/draft/2020-12/schema"),
            Dialect.from_str("https://json-schema.org/draft/2019-09/schema"),
            Dialect.from_str("http://json-schema.org/draft-07/schema#"),
            Dialect.from_str("http://json-schema.org/draft-06/schema#"),
            Dialect.from_str("http://json-schema.org/draft-04/schema#"),
            Dialect.from_str("http://json-schema.org/draft-03/schema#"),
        ],
        benchmarks=benchmarks,
        uri=URL.parse(Path(__file__).absolute().as_uri()),
        varying_parameter=varying_parameter,
    )



from __future__ import annotations

from collections import Counter
from collections.abc import Callable, Iterable
from contextlib import AsyncExitStack, asynccontextmanager
from fnmatch import fnmatch
from functools import wraps
from io import TextIOWrapper
from pathlib import Path
from pprint import pformat
from statistics import mean, median, quantiles
from textwrap import dedent
from time import perf_counter_ns
from typing import IO, TYPE_CHECKING, Any, Literal, ParamSpec, Protocol, cast
import asyncio
import json
import logging
import os
import sys
import tarfile

from click.shell_completion import CompletionItem
from diagnostic import DiagnosticError
from inflect import engine as InflectEngine
from rich import box, console, panel
from rich.progress import (
    BarColumn,
    DownloadColumn,
    MofNCompleteColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
)
from rich.table import Column, Table
from rich.text import Text
from rich_click.utils import CommandGroupDict, OptionGroupDict
from url import URL, RelativeURLWithoutBase
import httpx
import rich_click as click
import structlog
import structlog.typing

from bowtie import DOCS, HOMEPAGE, _benchmarks, _connectables, _report, _suite
from bowtie._commands import SeqCase, TestResult, Unsuccessful
from bowtie._core import (
    Dialect,
    Example,
    Implementation,
    Test,
    TestCase,
    convert_table_to_markdown,
)
from bowtie._direct_connectable import Direct
from bowtie.exceptions import (
    CannotConnect,
    DialectError,
    NoSuchImplementation,
    StartupFailed,
    UnsupportedDialect,
)

if TYPE_CHECKING:
    from collections.abc import (
        AsyncIterator,
        Awaitable,
        Mapping,
        Sequence,
        Set,
    )
    from os import PathLike
    from typing import TextIO

    from click.decorators import FC
    from httpx import Response
    from referencing.jsonschema import SchemaResource

    from bowtie._commands import AnyTestResult, SeqResult
    from bowtie._connectables import Connectable, ConnectableId
    from bowtie._core import DialectRunner, ImplementationInfo
    from bowtie._registry import ValidatorRegistry


class _EX:
    def __getattr__(self, attr: str) -> int:
        return getattr(os, f"EX_{attr}", 1)  # Windows fallbacks...


EX = _EX()

STDERR = console.Console(stderr=True)

STARTUP_ERRORS = (CannotConnect, NoSuchImplementation, StartupFailed)


# rich-click's CommandGroupDict seems to be missing some covariance,
# as using a regular dict here makes pyright complain.
_COMMAND_GROUPS = {
    "bowtie": [
        CommandGroupDict(
            name="Basic Commands",
            commands=["validate", "suite", "summary", "info"],
        ),
        CommandGroupDict(
            name="Advanced Usage",
            commands=[
                "filter-dialects",
                "filter-implementations",
                "latest-report",
                "run",
                "statistics",
                "trend",
            ],
        ),
        CommandGroupDict(
            name="Plumbing Commands",
            commands=["badges", "smoke"],
        ),
        CommandGroupDict(
            name="Benchmarking Commands",
            commands=["perf", "filter-benchmarks"],
        ),
    ],
}
_OPTION_GROUPS = {
    f"bowtie {command}": [
        *[
            OptionGroupDict(name=group, options=[f"--{o}" for o in options])
            for group, options in groups
        ],
        OptionGroupDict(
            name="Connection & Communication Options",
            options=["--read-timeout", "--validate-implementations"],
        ),
        OptionGroupDict(name="Help", options=["--help"]),
    ]
    for command, groups in [
        (
            "validate",
            [
                ("Required", ["implementation"]),
                ("Schema Behavior Options", ["dialect", "set-schema"]),
                ("Validation Metadata Options", ["description", "expect"]),
            ],
        ),
        (
            "suite",
            [
                ("Required", ["implementation"]),
                (
                    "Test Run Options",
                    ["fail-fast", "filter", "max-error", "max-fail"],
                ),
                ("Test Modification Options", ["set-schema"]),
            ],
        ),
        ("info", [("Basic Options", ["implementation", "format"])]),
        ("smoke", [("Basic Options", ["implementation", "quiet", "format"])]),
        (
            "filter-dialects",
            [
                ("Required", ["implementation"]),
                ("Filters", ["dialect", "latest", "boolean-schemas"]),
            ],
        ),
        (
            "filter-implementations",
            [
                (
                    "Filters",
                    [
                        "supports-dialect",
                        "language",
                        "direct",
                        "implementation",
                    ],
                ),
            ],
        ),
        (
            "run",
            [
                ("Required", ["implementation"]),
                ("Schema Behavior Options", ["dialect", "set-schema"]),
                (
                    "Test Run Options",
                    ["fail-fast", "filter", "max-error", "max-fail"],
                ),
            ],
        ),
        (
            "perf",
            [
                ("Required", ["implementation"]),
                (
                    "Benchmark Configuration Options",
                    [
                        "benchmark-file",
                        "dialect",
                        "keywords",
                        "loops",
                        "runs",
                        "test-suite",
                        "values",
                        "warmups",
                    ],
                ),
                (
                    "Basic Options",
                    ["format", "max-fail", "quiet"],
                ),
            ],
        ),
    ]
}


@click.rich_config(
    help_config=click.RichHelpConfiguration(
        command_groups=_COMMAND_GROUPS,
        option_groups=_OPTION_GROUPS,
        style_commands_table_column_width_ratio=(1, 3),
        # Otherwise there's an uncomfortable amount of internal whitespace.
        max_width=120,
    ),
)
@click.group(
    context_settings=dict(help_option_names=["--help", "-h"]),
    # needing to explicitly dedent here, as well as the extra newline
    # before "Full documentation" both seem like rich-click bugs.
    epilog=dedent(
        f"""
        If you don't know where to begin, `bowtie validate --help` or
        `bowtie suite --help` are likely good places to start.

        Full documentation can also be found at {DOCS}
        """,
    ),
)
@click.version_option(prog_name="bowtie", package_name="bowtie-json-schema")
@click.option(
    "--log-level",
    "-L",
    help="How verbose should Bowtie be?",
    default="warning",
    show_default="warning",
    type=click.Choice(
        [
            "debug",
            "info",
            "warning",
            "error",
            "critical",
        ],
    ),
)
def main(log_level: str):
    """
    A meta-validator for the JSON Schema specifications.

    Bowtie gives you access to the JSON Schema ecosystem across every
    programming language and implementation.

    It lets you compare implementations either to each other or to known
    correct results from the official JSON Schema test suite.
    """
    _redirect_structlog(log_level=getattr(logging, log_level.upper()))


P = ParamSpec("P")


def subcommand(fn: Callable[P, int | None]):
    """
    Define a Bowtie subcommand which returns its exit code.
    """

    @main.command()
    @click.pass_context
    @wraps(fn)
    def run(context: click.Context, *args: P.args, **kwargs: P.kwargs) -> None:
        exit_code = fn(*args, **kwargs)
        context.exit(0 if exit_code is None else exit_code)

    return run


class Starter(Protocol):

    connectables: Iterable[Connectable]

    def __call__(
        self,
        connectables: Iterable[Connectable] = [],
    ) -> AsyncIterator[tuple[ConnectableId, Implementation]]: ...


class ImplementationSubcommand(Protocol):
    def __call__(
        self,
        start: Starter,
        **kwargs: Any,
    ) -> Awaitable[int | None]: ...


SILENT = _report.Reporter(write=lambda **_: None)  # type: ignore[reportUnknownArgumentType])


def implementation_subcommand(
    reporter: _report.Reporter = SILENT,
    default_implementations: Set[str] = Implementation.known(),
):
    """
    Define a Bowtie subcommand which starts up some implementations.

    Runs the wrapped function with only the successfully started
    implementations.
    """

    def wrapper(fn: ImplementationSubcommand):
        async def run(
            connectables: Iterable[Connectable],
            registry: ValidatorRegistry[Any] = Direct.from_id(
                "python-jsonschema",
            ).registry(),
            **kw: Any,
        ) -> int:
            exit_code = 0

            async def start(
                connectables: Iterable[Connectable] = connectables,
            ):
                nonlocal exit_code

                successful = 0
                async with _start(
                    connectables=connectables,
                    registry=registry,
                    reporter=reporter,
                ) as implementations:
                    for each in implementations:  # FIXME: respect --quiet
                        try:
                            connectable_implementation = await each
                        except STARTUP_ERRORS as error:
                            exit_code |= EX.CONFIG
                            STDERR.print(error)
                            continue

                        successful += 1
                        yield connectable_implementation

                    if not successful and default_implementations:
                        # TODO: show a diagnostic that collects crash causes
                        exit_code |= EX.CONFIG
                        return

            # FIXME: Convert this to an instance presumably, but for now we
            #        just want this data available in the functions,
            #        and introducing another type is annoying when most of the
            #        complexity has to do with _run / _start still existing --
            #        we need to finish removing them.
            start.connectables = connectables  # type: ignore[reportFunctionMemberAccess]

            fn_exit_code = await fn(start=start, **kw)  # type: ignore[reportArgumentType]
            return exit_code | (fn_exit_code or 0)

        @subcommand
        @click.option(
            "--implementation",
            "-i",
            "connectables",
            type=_connectables.ClickParam(),
            default=lambda: (
                default_implementations
                if sys.stdin.isatty() or "CI" in os.environ
                else [line.strip() for line in sys.stdin]
            ),
            multiple=True,
            metavar="IMPLEMENTATION",
            help=(
                "A connectable ID for a JSON Schema implementation supported "
                "by Bowtie. May be repeated multiple times to select multiple "
                "implementations to run. "
                "Run `bowtie filter-implementations` for the full list of "
                "supported implementations."
            ),
        )
        @wraps(fn)
        def cmd(
            connectables: Iterable[Connectable],
            **kwargs: Any,
        ) -> int:
            return asyncio.run(run(connectables=connectables, **kwargs))

        return cmd

    return wrapper


@subcommand
@click.option(
    "--site",
    default=Path("site"),
    show_default=True,
    type=click.Path(
        path_type=Path,
        file_okay=False,
        dir_okay=True,
        exists=True,
    ),
    help=(
        "The path to a previously generated collection of reports, "
        "used to generate the badges."
    ),
)
def badges(site: Path):
    """
    Generate Bowtie badges for implementations using a previous Bowtie run.

    Will generate badges for any existing dialects, and ignore any for which a
    report was not generated.
    """
    outdir = site / "badges"
    try:
        outdir.mkdir()
    except FileExistsError:
        error = DiagnosticError(
            code="already-exists",
            message="Badge output directory already exists.",
            causes=[f"{outdir} is an existing directory."],
            hint_stmt=(
                "If you intended to replace its contents with new badges, "
                "delete the directory first."
            ),
        )
        STDERR.print(error)
        return EX.CONFIG

    supported_versions: dict[Path, Iterable[Dialect]] = {}

    for name, dialect in Dialect.by_short_name().items():
        try:
            file = site.joinpath(f"{name}.json").open()
        except FileNotFoundError:
            continue
        with file:
            report = _report.Report.from_serialized(file)
            if report.is_empty:
                error = DiagnosticError(
                    code="empty-report",
                    message="A Bowtie report is empty.",
                    causes=[f"The {name} report contains no results."],
                    hint_stmt="Check that site generation has not failed.",
                )
                STDERR.print(error)
                return EX.DATAERR

            badge_name = f"{dialect.short_name}.json"

            for each, badge in report.compliance_badges():
                dir = outdir / each.id

                compliance = dir / "compliance"
                compliance.mkdir(parents=True, exist_ok=True)
                compliance.joinpath(badge_name).write_text(json.dumps(badge))

                dialects = each.dialects
                seen = supported_versions.setdefault(dir, dialects)
                if seen != dialects:
                    message = (
                        f"{dir.name} appears with different "
                        "supported dialects in the provided reports."
                    )
                    error = DiagnosticError(
                        code="inconsistent-reports",
                        message=message,
                        causes=[
                            f"{file.name} contains:\n{pformat(dialects)}",
                            f"{pformat(seen)} was previously seen.",
                        ],
                        hint_stmt=(
                            "Check that the implementation produces "
                            "consistent output and that a run has not failed."
                        ),
                    )
                    STDERR.print(error)
                    return EX.CONFIG

    for dir, dialects in supported_versions.items():
        badge = _report.supported_version_badge(dialects=dialects)
        dir.joinpath("supported_versions.json").write_text(json.dumps(badge))


_F = Literal["json", "pretty", "markdown"]


def format_option(**option_kwargs: Any) -> Callable[[FC], FC]:
    if not option_kwargs:
        option_kwargs = dict(
            default=lambda: "pretty" if sys.stdout.isatty() else "json",
            show_default="pretty if stdout is a tty, otherwise JSON",
            type=click.Choice(["json", "pretty", "markdown"]),
        )

    def _format_option(fn: FC) -> FC:
        def show_schema(
            ctx: click.Context,
            param: click.Parameter | None,
            value: bool,
        ) -> None:
            if not value or ctx.resilient_parsing:
                return
            uri = URL.parse(f"tag:bowtie.report,2024:cli:{ctx.command.name}")
            schema = Direct.from_id("python-jsonschema").registry().schema(uri)
            # FIXME: Syntax highlight? But rich appears to be doing some
            #        bizarre line wrapping, even if I disable a bunch of random
            #        options (crop, no_wrap, word_wrap in Syntax, ...) which
            #        fails the integration tests.
            click.echo(json.dumps(schema, indent=2))
            ctx.exit()

        return click.option(
            "--format",
            "-f",
            "format",
            help="What format to use for the output",
            **option_kwargs,
        )(
            click.option(
                "--schema",
                callback=show_schema,
                expose_value=False,
                is_eager=True,
                is_flag=True,
                help="Show the JSON Schema for this command's JSON output.",
            )(fn),
        )

    return _format_option


class _Report(click.File):
    """
    Select a previously produced Bowtie report.
    """

    name = "report"
    mode = "r"

    def convert(  # type: ignore[reportIncompatibleMethodOverride]
        self,
        value: str | PathLike[str] | IO[Any] | _report.Report,
        param: click.Parameter | None,
        ctx: click.Context,
    ) -> _report.Report:
        if isinstance(value, _report.Report):
            return value

        input = super().convert(value, param, ctx)
        try:
            return _report.Report.from_serialized(input)
        except _report.EmptyReport:
            error = DiagnosticError(
                code="empty-report",
                message="The Bowtie report is empty.",
                causes=[f"{input.name} contains no test result data."],
                hint_stmt=(
                    "If you are piping in report data, "
                    "check to ensure that what you've run has succeeded, "
                    "otherwise it may be emitting no report data."
                ),
            )
            STDERR.print(error)
            ctx.exit(EX.NOINPUT)
        except json.JSONDecodeError as err:
            error = DiagnosticError(
                code="report-not-json",
                message="The Bowtie report looks corrupt.",
                causes=[f"{input.name} is not valid JSON.", str(err)],
                hint_stmt=(
                    "If you are piping in report data, "
                    "the command producing the report has likely failed "
                    "and the real error is above this one. "
                    "Otherwise, ensure you are passing in a report generated "
                    "by Bowtie."
                ),
            )
            STDERR.print(error)
            ctx.exit(EX.DATAERR)
        except _report.MissingFooter:
            error = DiagnosticError(
                code="truncated-report",
                message="The Bowtie report looks corrupt.",
                causes=[
                    f"{input.name} is missing its footer, which usually means "
                    "it has been somehow truncated.",
                ],
                hint_stmt=(
                    "Try running the command you used to produce the report, "
                    "without piping it. If it crashes, file a bug report!"
                ),
            )
            STDERR.print(error)
            ctx.exit(EX.DATAERR)


@subcommand
@format_option()
@click.option(
    "--show",
    "-s",
    default="validation",
    show_default=True,
    type=click.Choice(["failures", "validation"]),
    help=(
        "Configure whether to display validation results "
        "(whether instances are valid or not) or test failure results "
        "(whether the validation results match expected validation results)"
    ),
)
@click.argument("report", default="-", type=_Report())
def summary(report: _report.Report, format: _F, show: str):
    """
    Generate an (in-terminal) summary of a Bowtie run.
    """
    if show == "failures":
        results = report.worst_to_best()
        exit_code = (
            EX.DATAERR
            if any(
                unsuccessful.failed or unsuccessful.errored
                for _, __, unsuccessful in results
            )
            else 0
        )
        to_table = _failure_table
        to_markdown_table = _failure_table_in_markdown

        def to_serializable(  # type: ignore[reportRedeclaration]
            value: Iterable[
                tuple[ConnectableId, ImplementationInfo, Unsuccessful],
            ],
        ):
            return [(id, u.counts()) for id, _, u in value]

    else:
        results = report.cases_with_results()
        exit_code = 0
        to_table = _validation_results_table
        to_markdown_table = _validation_results_table_in_markdown

        def to_serializable(
            value: Iterable[
                tuple[
                    TestCase,
                    Iterable[tuple[Test, dict[str, AnyTestResult]]],
                ]
            ],
        ):
            return [
                (
                    case.schema,
                    [
                        (
                            test.instance,
                            {k: v.description for k, v in test_result.items()},
                        )
                        for test, test_result in test_results
                    ],
                )
                for case, test_results in value
            ]

    match format:
        case "json":
            click.echo(json.dumps(to_serializable(results), indent=2))  # type: ignore[reportGeneralTypeIssues]
        case "pretty":
            table = to_table(report, results)  # type: ignore[reportGeneralTypeIssues]
            console.Console().print(table)
        case "markdown":
            table = to_markdown_table(report, results)  # type: ignore[reportGeneralTypeIssues]
            console.Console().print(table)

    return exit_code


def _failure_table(
    report: _report.Report,
    results: list[tuple[ConnectableId, ImplementationInfo, Unsuccessful]],
):
    test = "tests" if report.total_tests != 1 else "test"
    table = Table(
        "Implementation",
        "Skips",
        "Errors",
        "Failures",
        title="Bowtie",
        caption=f"{report.total_tests} {test} ran\n",
    )

    implementation_counts = Counter(
        each.id for each in report.implementations.values()
    )
    for _, each, unsuccessful in results:
        table.add_row(
            Text.assemble(
                each.name,
                (
                    (f" {each.version}", "dim")
                    if implementation_counts[each.id] > 1
                    else ("", "")
                ),
                (f" ({each.language})", "dim"),
            ),
            str(len(unsuccessful.skipped)),
            str(len(unsuccessful.errored)),
            str(len(unsuccessful.failed)),
        )
    return table


def _failure_table_in_markdown(
    report: _report.Report,
    results: list[tuple[ConnectableId, ImplementationInfo, Unsuccessful]],
):
    test = "tests" if report.total_tests != 1 else "test"
    rows: list[list[str]] = []
    columns = [
        "Implementation",
        "Skips",
        "Errors",
        "Failures",
    ]

    implementation_counts = Counter(
        each.id for each in report.implementations.values()
    )
    for _, each, unsuccessful in results:
        rows.append(
            [
                f"{each.name}"
                + (
                    f" {each.version}"
                    if implementation_counts[each.id] > 1
                    else ""
                )
                + f" ({each.language})",
                str(len(unsuccessful.skipped)),
                str(len(unsuccessful.errored)),
                str(len(unsuccessful.failed)),
            ],
        )

    return "\n".join(
        [
            "# Bowtie Failures Summary",
            convert_table_to_markdown(columns, rows),
            "",
            f"**{report.total_tests} {test} ran**",
        ],
    )


def _validation_results_table(
    report: _report.Report,
    results: Iterable[
        tuple[TestCase, Iterable[tuple[Test, Mapping[str, AnyTestResult]]]],
    ],
):
    test = "tests" if report.total_tests != 1 else "test"
    table = Table(
        Column(header="Schema", vertical="middle"),
        "",
        title="Bowtie",
        caption=f"{report.total_tests} {test} ran",
    )

    # TODO: sort the columns by results?
    implementations = report.implementations
    implementation_counts = Counter(
        each.id for each in implementations.values()
    )

    for case, test_results in results:
        subtable = Table("Instance", box=box.SIMPLE_HEAD)
        for implementation in implementations.values():
            subtable.add_column(
                Text.assemble(
                    implementation.name,
                    (
                        (f" {implementation.version}", "dim")
                        if implementation_counts[implementation.id] > 1
                        else ("", "")
                    ),
                    (f" ({implementation.language})", "dim"),
                ),
            )

        for test, test_result in test_results:
            subtable.add_row(
                test.syntax(),
                *(Text(test_result[id].description) for id in implementations),
            )

        table.add_row(case.syntax(report.metadata.dialect), subtable)
        table.add_section()

    return table


def _validation_results_table_in_markdown(
    report: _report.Report,
    results: Iterable[
        tuple[TestCase, Iterable[tuple[Test, Mapping[str, AnyTestResult]]]],
    ],
):
    rows_data: list[list[str]] = []
    final_content = ""

    inner_table_columns = ["Instance"]
    implementations = report.implementations
    implementation_counts = Counter(
        each.id for each in implementations.values()
    )
    inner_table_columns.extend(
        f"{implementation.name}"
        + (
            f" {implementation.version}"
            if implementation_counts[implementation.id] > 1
            else ""
        )
        + f" ({implementation.language})"
        for implementation in implementations.values()
    )

    for case, test_results in results:
        inner_table_rows: list[list[str]] = []
        for test, test_result in test_results:
            inner_table_rows.append(
                [
                    json.dumps(test.instance),
                    *(test_result[id].description for id in implementations),
                ],
            )
        inner_markdown_table = convert_table_to_markdown(
            inner_table_columns,
            inner_table_rows,
        )
        schema_name = json.dumps(case.schema, indent=2)
        row_data = [schema_name, inner_markdown_table]
        rows_data.append(row_data)

    for idx, row_data in enumerate(rows_data):
        final_content += f"### {idx+1}. Schema:\n {row_data[0]}\n\n"
        final_content += "### Results:"
        final_content += row_data[1]

    return final_content


@subcommand
@format_option()
@click.option(
    "--quantiles",
    "n",
    default=4,
    type=int,
    help=(
        "How many quantiles should be emitted for the compliance numbers? "
        "Computing quantiles only is sensical if this number is more than the "
        "number of implementations reported on. By default, we compute "
        "quartiles."
    ),
)
@click.argument(
    "report",
    default=lambda: (
        "-"
        if not sys.stdin.isatty()
        else _report.Report.from_serialized(
            asyncio.run(Dialect.latest().latest_report()).iter_lines(),
        )
    ),
    type=_Report(),
)
def statistics(
    report: _report.Report,
    n: int,
    format: _F,
):
    """
    Show summary statistics for a Bowtie generated report.

    If stdin is a TTY, the most recent public report for the latest JSON Schema
    dialect is downloaded.
    Otherwise, if it *is not* a TTY (e.g. if it is a pipe) then it should
    contain report data.

    Piping input via:

      $ bowtie latest-report --dialect <some-dialect> | bowtie statistics

    can be useful to retrieve the latest report for a specific dialect.
    """
    dialect, ran_on_date = report.metadata.dialect, report.metadata.started
    unsuccessful = report.compliance_by_implementation().values()
    statistics = dict(
        median=median(unsuccessful),
        mean=mean(unsuccessful),
        **(  # quantiles only make sense for n < len(data)
            {"quantiles": quantiles(unsuccessful, n=n)}
            if n < len(unsuccessful)
            else {}
        ),
    )
    match format:
        case "json":
            statistics = {
                "dialect": str(dialect.uri),
                "ran_on": ran_on_date.isoformat(),
                **statistics,
            }
            click.echo(json.dumps(statistics, indent=2))
        case "pretty":
            click.echo(
                f"Dialect: {dialect.pretty_name}\n"
                f"Ran on: {ran_on_date.strftime('%x %X %Z')}\n",
            )
            for k, v in statistics.items():
                click.echo(f"{k}: {v}")
        case "markdown":
            heading = (
                f"## Dialect: {dialect.pretty_name}\n\n"
                f"### Ran on: {ran_on_date.strftime('%x %X %Z')}\n"
            )
            markdown = convert_table_to_markdown(
                columns=["Metric", "Value"],
                rows=[[k, str(v)] for k, v in statistics.items()],
            )
            click.echo(heading + markdown)


def do_not_validate(*ignored: SchemaResource) -> Callable[..., None]:
    return lambda *args, **kwargs: None


class _Dialect(click.ParamType):
    """
    Select a JSON Schema dialect.
    """

    name = "dialect"

    def convert(
        self,
        value: str | Dialect,
        param: click.Parameter | None,
        ctx: click.Context | None,
    ) -> Dialect:
        if not isinstance(value, str):
            return value

        dialect = Dialect.by_alias().get(value)
        if dialect is not None:
            return dialect

        try:
            url = URL.parse(value)
        except RelativeURLWithoutBase:
            pass
        else:
            dialect = Dialect.by_uri().get(url)
            if dialect is not None:
                return dialect

        self.fail(f"{value!r} is not a known dialect URI or short name.")

    def shell_complete(
        self,
        ctx: click.Context,
        param: click.Parameter,
        incomplete: str,
    ) -> list[CompletionItem]:
        if incomplete:  # the user typed something, so filter over everything
            suggestions = [
                (field, dialect)
                for dialect in Dialect.known()
                for field in [
                    str(dialect.uri),
                    dialect.short_name,
                    *dialect.aliases,
                ]
            ]
        else:  # the user didn't type anything, only suggest short names
            suggestions = Dialect.by_short_name().items()

        return [
            # FIXME: pallets/click#2703
            CompletionItem(
                value=value.replace(":", "\\:"),
                help=f"the {dialect.pretty_name} dialect",
            )
            for value, dialect in suggestions
            if value.startswith(incomplete.lower())
        ]


CaseTransform = Callable[[Iterable[TestCase]], Iterable[TestCase]]


class _Filter(click.ParamType):
    """
    Filter some test cases by a pattern.
    """

    name = "filter"

    def convert(
        self,
        value: str,
        param: click.Parameter | None,
        ctx: click.Context | None,
    ) -> CaseTransform:
        return lambda cases: (
            case for case in cases if fnmatch(case.description, f"*{value}*")
        )


def _set_dialect_via_schema(ctx: click.Context, _, value: _Dialect):
    """
    Set the dialect according to a possibly present :kw:`$schema` keyword.
    """
    if value:
        return value
    schema = ctx.params.get("schema")
    dialect_from_schema: str | None = (  # type: ignore[reportUnknownVariableType]
        schema.get("$schema")  # type: ignore[reportUnknownMemberType]
        if isinstance(schema, dict)
        else None
    )
    return (
        Dialect.from_str(dialect_from_schema)  # type: ignore[reportUnknownArgumentType]
        if dialect_from_schema
        else Dialect.latest()
    )


def _set_schema(dialect: Dialect) -> CaseTransform:
    """
    Explicitly set a dialect on schemas passing through by setting ``$schema``.
    """
    return lambda cases: (c.with_explicit_dialect(dialect) for c in cases)


def _do_nothing(*args: Any, **kwargs: Any) -> CaseTransform:
    return lambda cases: cases


IMPLEMENTATION = click.option(
    "--implementation",
    "-i",
    "connectables",
    type=_connectables.ClickParam(),
    required=True,
    multiple=True,
    metavar="IMPLEMENTATION",
    help=(
        "A connectable ID for a JSON Schema implementation supported "
        "by Bowtie. May be repeated multiple times to select multiple "
        "implementations to run. "
        "Run `bowtie filter-implementations` for the full list of "
        "supported implementations."
    ),
)
FILTER = click.option(
    "--filter",
    "-k",
    default="",
    type=_Filter(),
    metavar="GLOB",
    help="Only run cases whose description match the given glob pattern.",
)
SET_SCHEMA = click.option(
    "--set-schema",
    "-S",
    "maybe_set_schema",
    # I have no idea why Click makes this so hard, but no combination of:
    #     type, default, is_flag, flag_value, nargs, ...
    # makes this work without doing it manually with callback.
    callback=lambda _, __, v: _set_schema if v else _do_nothing,  # type: ignore[reportUnknownLambdaType]
    is_flag=True,
    show_default=True,
    default=False,
    help=(
        "Explicitly set $schema in all (non-boolean) case schemas sent to "
        "implementations. Note this of course means what is passed to "
        "implementations will differ from what is provided in the input."
    ),
)
VALIDATE = click.option(
    "--validate-implementations",
    "-V",
    "registry",
    # I have no idea why Click makes this so hard, but no combination of:
    #     type, default, is_flag, flag_value, nargs, ...
    # makes this work without doing it manually with callback.
    callback=lambda _, __, v: (  # type: ignore[reportUnknownLambdaType]
        Direct.from_id("python-jsonschema" if v else "null").registry()
    ),
    is_flag=True,
    help=(
        "When speaking to implementations (provided via -i), validate "
        "the requests and responses sent to them under Bowtie's JSON Schema "
        "specification. Generally, this option protects against broken Bowtie "
        "implementations and can be left at its default (of off) unless "
        "you are developing a new implementation container."
    ),
)

_inflect_engine = InflectEngine()

POSSIBLE_DIALECT_SHORTNAMES = _inflect_engine.join(sorted(Dialect.by_alias()))  # type: ignore[reportArgumentType]


def pretty_names_str_for(dialects: Iterable[Dialect]) -> str:
    return (
        "all known dialects"
        if list(dialects) == list(Dialect.known())
        else _inflect_engine.join(
            [dialect.pretty_name for dialect in dialects],  # type: ignore[reportArgumentType]
        )
    )


def dialect_option(
    default: Dialect | None = Dialect.latest(),
    **kwargs: Any,
):
    if default is not None:
        kwargs.update(default=default, show_default=default.pretty_name)

    return click.option(
        "--dialect",
        "-D",
        "dialect",
        type=_Dialect(),
        metavar="URI_OR_NAME",
        help=(
            "A URI or shortname identifying the dialect of each test. "
            f"Possible shortnames include: {POSSIBLE_DIALECT_SHORTNAMES}."
        ),
        **kwargs,
    )


def fail_fast(fn: FC) -> FC:
    conflict = "don't provide both --fail-fast and --max-fail / --max-error"

    # Both are these are needed because parsing is order dependent :/
    def disallow_fail_fast(
        ctx: click.Context,
        _,
        value: int | None,
    ) -> int | None:
        if ctx.params.get("fail_fast"):
            if value is None:
                return 1
            raise click.UsageError(conflict)
        return value

    def disallow_max_fail(
        ctx: click.Context,
        _,
        value: int | None,
    ) -> int | None:
        if value and ctx.params.get("max_fail", 1) != 1:
            raise click.UsageError(conflict)
        return value

    N = "COUNT"
    msg = f"Stop running once {N} tests {{}} in total across implementations."
    return click.option(
        "-x",
        "--fail-fast",
        callback=disallow_max_fail,
        is_flag=True,
        default=False,
        help="Stop running immediately after the first failure or error.",
    )(
        click.option(
            "--max-fail",
            metavar=N,
            type=click.IntRange(min=1),
            callback=disallow_fail_fast,
            help=msg.format("fail"),
        )(
            click.option(
                "--max-error",
                metavar=N,
                type=click.IntRange(min=1),
                callback=disallow_fail_fast,
                help=msg.format("error"),
            )(fn),
        ),
    )


class JSON(click.File):

    name = "JSON"

    def convert(
        self,
        value: str | PathLike[str] | IO[Any],
        param: click.Parameter | None,
        ctx: click.Context | None,
    ) -> Any:
        return json.load(super().convert(value, param, ctx))


@subcommand
@dialect_option()
@IMPLEMENTATION
@FILTER
@fail_fast
@SET_SCHEMA
@VALIDATE
@click.argument(
    "input",
    default="-",
    type=click.File(mode="rb"),
)
def run(
    input: Iterable[str],
    filter: CaseTransform,
    dialect: Dialect,
    **kwargs: Any,
):
    """
    Run test cases written directly in Bowtie's testing format.

    This is generally useful if you wish to hand-author which schemas to
    include in the schema registry, or otherwise exactly control the contents
    of a test case.
    """
    cases = filter(
        TestCase.from_dict(dialect=dialect, **json.loads(line))
        for line in input
    )

    return asyncio.run(_run(**kwargs, cases=cases, dialect=dialect))


@subcommand
@dialect_option(default=None, callback=_set_dialect_via_schema)
@IMPLEMENTATION
@SET_SCHEMA
@VALIDATE
@click.option(
    "-d",
    "--description",
    default="bowtie validate",
    help="A (human-readable) description for this test case.",
)
@click.option(
    "--expect",
    show_default=True,
    show_choices=True,
    default="any",
    type=click.Choice(["valid", "invalid", "any"], case_sensitive=False),
    callback=lambda _, __, value: (  # type: ignore[reportUnknownLambdaType]
        None if value == "any" else value == "valid"
    ),
    help=(
        "Expect the given input to be considered valid or invalid, "
        "or else (with 'any') to allow either result."
    ),
)
@click.argument("schema", type=JSON())
@click.argument("instances", nargs=-1, type=JSON())
def validate(
    schema: Any,
    instances: Iterable[Any],
    expect: bool | None,
    description: str,
    **kwargs: Any,
):
    """
    Validate instances under a schema across any supported implementation.
    """
    if not instances:
        return EX.NOINPUT

    tests = [Example(description="", instance=each) for each in instances]
    if expect is not None:
        tests = [test.expect(expect) for test in tests]
    case = TestCase(description=description, schema=schema, tests=tests)
    return asyncio.run(_run(fail_fast=False, **kwargs, cases=[case]))


def _set_benchmarker_callable(
    ctx: click.Context,
    value: Any,
    callable: Callable[..., Any],
) -> Any:
    if value:
        ctx.params["benchmarker_callable"] = callable
    return value


@subcommand
@IMPLEMENTATION
@dialect_option()
@format_option()
@click.option(
    "--runs",
    "-r",
    "runs",
    type=int,
    default=3,
    show_default=True,
    help="Number of runs used to run benchmarks.",
)
@click.option(
    "--values",
    "-v",
    "values",
    type=click.IntRange(min=2),
    default=2,
    show_default=True,
    help="Number of values per run.",
)
@click.option(
    "--warmups",
    "-w",
    "warmups",
    type=int,
    default=1,
    show_default=True,
    help="Number of skipped values per run used to warmup the benchmark.",
)
@click.option(
    "--loops",
    "-l",
    "loops",
    type=int,
    default=1,
    show_default=True,
    help="Number of loops per value.",
)
@click.option(
    "--quiet",
    "-q",
    "quiet",
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable quiet mode (Only output the final result).",
)
@click.option(
    "--keywords",
    "-k",
    "keywords",
    callback=lambda ctx, __, value: (  # type: ignore[reportUnknownLambdaType]
        _set_benchmarker_callable(
            ctx,  # type: ignore[reportUnknownArgumentType]
            value,
            _benchmarks.Benchmarker.for_keywords,
        )
    ),
    is_flag=True,
    show_default=True,
    help=(
        "Run keyword specific benchmarks to learn about how "
        "various implementations implement the keyword."
    ),
)
@click.option(
    "-b",
    "--benchmark-file",
    "benchmark_files",
    callback=lambda ctx, __, value: (  # type: ignore[reportUnknownLambdaType]
        _set_benchmarker_callable(
            ctx,  # type: ignore[reportUnknownArgumentType]
            value,
            _benchmarks.Benchmarker.for_benchmark_files,
        )
    ),
    multiple=True,
    help=(
        "Allows running benchmark from a file. "
        "Specify the path of the benchmark file"
    ),
)
@click.option(
    "--test-suite",
    "-t",
    "test_suite",
    callback=lambda ctx, __, value: (  # type: ignore[reportUnknownLambdaType]
        _set_benchmarker_callable(
            ctx,  # type: ignore[reportUnknownArgumentType]
            value,
            _benchmarks.Benchmarker.from_test_cases,
        )
    ),
    type=_suite.ClickParam(),
    default=None,
    help="Run Benchmarks over the official JSON Schema Test Suite.",
)
@click.argument(
    "benchmark",
    type=JSON(),
    required=False,
    callback=lambda ctx, __, value: (  # type: ignore[reportUnknownLambdaType]
        _set_benchmarker_callable(
            ctx,  # type: ignore[reportUnknownArgumentType]
            value,
            _benchmarks.Benchmarker.from_input,
        )
    ),
)
def perf(
    connectables: Iterable[_connectables.Connectable],
    dialect: Dialect,
    format: _F,
    quiet: bool,
    benchmarker_callable: Callable[..., Any] = (
        _benchmarks.Benchmarker.from_default_benchmarks
    ),
    **kwargs: Any,
):
    """
    Perform performance measurements across supported implementations.
    """
    if kwargs.get("test_suite"):
        cases, enforced_dialect, _ = kwargs["test_suite"]
        dialect = enforced_dialect
        kwargs["cases"] = cases

    try:
        benchmarker = benchmarker_callable(dialect=dialect, **kwargs)
        asyncio.run(
            benchmarker.start(
                connectables=connectables,
                dialect=dialect,
                quiet=quiet,
                format=format,
            ),
        )
    except (_benchmarks.BenchmarkError, _benchmarks.BenchmarkLoadError) as err:
        STDERR.print(err)
        return EX.DATAERR

    return 0


@subcommand
@dialect_option()
@click.option(
    "-t",
    "--benchmark-type",
    type=click.Choice(["default", "keyword"]),
    default=None,
    show_default=True,
    help=("Specify the type of benchmark to filter."),
)
@click.option(
    "-n",
    "--name",
    "benchmark_names",
    type=str,
    multiple=True,
    help=(
        "Filter the benchmarks with given name. "
        "Use the option multiple times to filter multiple benchmarks."
    ),
)
def filter_benchmarks(
    dialect: Dialect,
    benchmark_type: str | None,
    benchmark_names: Iterable[str],
):
    """
    Output benchmarks matching the specified criteria.
    """
    files: list[Path] = _benchmarks.get_benchmark_files(
        benchmark_type,
        benchmarks=benchmark_names,
        dialect=dialect,
    )
    for file in files:
        console.Console().file.write(f"{file}\n")


LANGUAGE_ALIASES = {
    "cpp": "c++",
    "js": "javascript",
    "ts": "typescript",
}
KNOWN_LANGUAGES = {
    *LANGUAGE_ALIASES.values(),
    *(i.partition("-")[0] for i in Implementation.known()),
}


@implementation_subcommand()  # type: ignore[reportArgumentType]
@format_option(
    default="plain",
    show_default=True,
    type=click.Choice(["plain", "json"]),
)
@click.option(
    "--supports-dialect",
    "-d",
    "dialects",
    type=_Dialect(),
    default=frozenset(),
    metavar="URI_OR_NAME",
    multiple=True,
    help=(
        "Only include implementations supporting the given dialect URI "
        "or dialect shortname. "
        f"Possible shortnames include: {POSSIBLE_DIALECT_SHORTNAMES}."
    ),
)
@click.option(
    "--language",
    "-l",
    "languages",
    type=click.Choice(sorted(KNOWN_LANGUAGES), case_sensitive=False),
    callback=lambda _, __, value: (  # type: ignore[reportUnknownLambdaType]
        KNOWN_LANGUAGES
        if not value
        else frozenset(
            LANGUAGE_ALIASES.get(each, each)  # type: ignore[reportUnknownArgumentType]
            for each in value  # type: ignore[reportUnknownArgumentType]
        )
    ),
    multiple=True,
    metavar="LANGUAGE",
    help="Only include implementations in the given programming language.",
)
@click.option(
    "--direct",
    "filter_connectable",
    is_flag=True,
    callback=lambda _, __, value: (  # type: ignore[reportUnknownLambdaType]
        (lambda connectable: connectable.kind == "direct")  # type: ignore[reportUnknownLambdaType]
        if value
        else (lambda connectable: True)  # type: ignore[reportUnknownLambdaType]
    ),
    help=(
        "Only include implementations with direct connectable functionality "
        "(i.e. which can run without the presence of a container runtime)."
    ),
)
async def filter_implementations(
    start: Starter,
    dialects: Sequence[Dialect],
    languages: Set[str],
    filter_connectable: Callable[[Connectable], bool],
    format: Literal["plain", "json"],
):
    """
    Output implementations which match the given criteria.

    Useful for piping or otherwise using the resulting output for further
    Bowtie commands.
    """
    filtered = (c for c in start.connectables if filter_connectable(c))
    if not dialects and languages == KNOWN_LANGUAGES:
        # speed up `bowtie filter-implementations`with no args or with --direct
        matching = [each.to_terse() for each in filtered]
    else:
        matching: Sequence[str] = [
            name
            async for name, each in start(connectables=filtered)
            if each.supports(*dialects) and each.info.language in languages
        ]

    match format:
        case "json":
            click.echo(json.dumps(matching, indent=2))
        case "plain":
            for name in matching:
                click.echo(name)


@implementation_subcommand(default_implementations=frozenset())  # type: ignore[reportArgumentType]
@click.option(
    "--dialect",
    "-d",
    "dialects",
    type=_Dialect(),
    default=Dialect.known(),
    metavar="URI_OR_NAME",
    multiple=True,
    help="Filter from the given list of dialects only.",
)
@click.option(
    "--latest",
    "-l",
    "latest",
    is_flag=True,
    default=False,
    help="Show only the latest dialect.",
)
@click.option(
    "--boolean-schemas/--no-boolean-schemas",
    "-b/-B",
    "booleans",
    default=None,
    help=(
        "If provided, show only dialects which do (or do not) "
        "support boolean schemas. Otherwise show either kind."
    ),
)
async def filter_dialects(
    start: Starter,
    dialects: Iterable[Dialect],
    latest: bool,
    booleans: bool | None,
):
    """
    Output dialect URIs matching a given criteria.

    If any implementations are provided, filter dialects supported by all the
    given implementations.
    """
    matching = {
        dialect
        for dialect in dialects
        if booleans is None or dialect.has_boolean_schemas == booleans
    }

    async for _, implementation in start():
        matching &= implementation.info.dialects

    if not matching:
        click.echo("No dialects match.", file=sys.stderr)
        return EX.DATAERR

    for dialect in sorted(matching, reverse=True):
        click.echo(dialect.uri)
        if latest:
            break


@subcommand
@dialect_option()
def latest_report(dialect: Dialect):
    """
    Output the latest published report from Bowtie's website.
    """

    async def write(response: Awaitable[Response]):
        async for chunk in (await response).aiter_bytes():
            click.echo(chunk)

    asyncio.run(write(dialect.latest_report()))


@implementation_subcommand()  # type: ignore[reportArgumentType]
@format_option()
@click.option(
    "--versions/--no-versions",
    "show_versions",
    help="Also show all implementation versions which Bowtie supports.",
)
async def info(
    start: Starter,
    format: _F,
    show_versions: bool,
):
    """
    Show information about a supported implementation.
    """
    serializable: dict[ConnectableId, dict[str, Any]] = {}

    async for _, each in start():
        metadata = [(k, v) for k, v in each.info.serializable().items() if v]
        if show_versions:
            metadata.append(("versions", await each.get_versions()))

        metadata.sort(
            key=lambda kv: (
                kv[0] != "name",
                kv[0] != "language",
                kv[0] != "version",
                kv[0] == "links",
                kv[0] == "versions",
                kv[0] == "dialects",
                kv[0],
            ),
        )

        match format:
            case "json":
                serializable[each.id] = dict(metadata)
            case "pretty":
                click.echo(
                    "\n".join(
                        f"{k}: {json.dumps(v, indent=2)}" for k, v in metadata
                    ),
                )
            case "markdown":
                click.echo(
                    "\n".join(
                        f"**{k}**: {json.dumps(v, indent=2)}"
                        for k, v in metadata
                    ),
                )

    if format == "json":
        if len(serializable) == 1:
            (output,) = serializable.values()
        else:
            output = serializable
        click.echo(json.dumps(output, indent=2))


async def download_versions_of(id: ConnectableId) -> frozenset[str]:
    progress = Progress(
        TextColumn("[bold blue]{task.description}"),
        SpinnerColumn(finished_text=""),
        BarColumn(bar_width=None),
        TextColumn(
            "[progress.percentage]{task.percentage:>3.0f}%",
        ),
        "•",
        DownloadColumn(),
        "•",
        TimeElapsedColumn(),
        console=console.Console(),
        transient=True,
    )
    task = progress.add_task(
        description=f"Fetching versions of {id}",
        total=None,
    )

    async with httpx.AsyncClient(timeout=10) as client:
        with progress:
            try:
                url = (
                    HOMEPAGE / "implementations" / id / "matrix-versions.json"
                )
                response = await client.get(str(url))
                response.raise_for_status()
            except httpx.HTTPStatusError as err:
                progress.update(
                    task,
                    description=(
                        f"[bold red]Could not fetch versions of "
                        f"{id}: {err.response.status_code}"
                    ),
                    completed=None,
                    total=None,
                    advance=None,
                )
                return frozenset()
            else:
                content = response.content
                content_length = len(content)

                progress.update(
                    task,
                    description=(
                        f"Successfully fetched all versions of {id}!"
                    ),
                    completed=content_length,
                    total=content_length,
                    advance=content_length,
                )
                return frozenset(json.loads(content))


async def download_and_parse_reports_for(
    id: ConnectableId,
    versions: Set[str],
    dialects: Iterable[Dialect],
) -> Iterable[tuple[str, Dialect, _report.Report]]:
    pretty_names_str = pretty_names_str_for(dialects)

    progress = Progress(
        TextColumn("[bold blue]{task.description}"),
        SpinnerColumn(finished_text=""),
        BarColumn(bar_width=None),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        "•",
        MofNCompleteColumn(),
        "•",
        TimeElapsedColumn(),
        console=console.Console(),
        transient=True,
    )

    if versions:
        total_files = len(versions) * len(list(dialects))
        actual_downloaded_files = 0

        task = progress.add_task(
            description=(
                "Preparing to download and parse versioned "
                f"reports of {id} for {pretty_names_str}"
            ),
            total=total_files,
        )

        async with httpx.AsyncClient(timeout=10) as client:

            async def download_and_parse_versioned_report_for(
                version: str,
                dialect: Dialect,
            ):
                try:
                    url = (
                        HOMEPAGE
                        / "implementations"
                        / id
                        / f"v{version}"
                        / f"{dialect.short_name}.json"
                    )
                    response = await client.get(str(url))
                    response.raise_for_status()
                    progress.update(
                        task,
                        description=(
                            f"Downloading and Parsing: "
                            f"v{version}/{dialect.short_name}.json"
                        ),
                    )
                except httpx.HTTPStatusError:
                    report = _report.Report.empty(dialect=dialect)
                    progress.update(task, advance=1)
                    return version, dialect, report
                else:
                    nonlocal actual_downloaded_files
                    actual_downloaded_files += 1

                    report = _report.Report.from_serialized(
                        response.iter_lines(),
                    )
                    progress.update(task, advance=1)
                    return version, dialect, report

            with progress:
                responses = await asyncio.gather(
                    *[
                        download_and_parse_versioned_report_for(
                            version,
                            dialect,
                        )
                        for version in versions
                        for dialect in dialects
                    ],
                )

                progress.update(
                    task,
                    description=(
                        "Successfully downloaded and parsed all versioned "
                        f"reports of {id} for {pretty_names_str}!"
                    ),
                    completed=actual_downloaded_files,
                    total=actual_downloaded_files,
                )
                return responses
    else:
        total_files = len(list(dialects))

        task = progress.add_task(
            description=(
                f"Preparing to download and parse latest "
                f"reports of {id} for {pretty_names_str}"
            ),
            total=total_files,
        )

        async def download_and_parse_latest_report_for(dialect: Dialect):
            try:
                response = await dialect.latest_report()
                response.raise_for_status()
                progress.update(
                    task,
                    description=(
                        f"Downloading and Parsing: "
                        f"latest/{dialect.short_name}.json"
                    ),
                )
            except httpx.HTTPStatusError:
                report = _report.Report.empty(dialect=dialect)
                progress.update(task, advance=1)
                return "latest", dialect, report
            else:
                report = _report.Report.from_serialized(response.iter_lines())
                progress.update(task, advance=1)
                return "latest", dialect, report

        with progress:
            responses = await asyncio.gather(
                *[
                    download_and_parse_latest_report_for(dialect)
                    for dialect in dialects
                ],
            )

            progress.update(
                task,
                description=(
                    f"Successfully downloaded and parsed all latest "
                    f"reports of {id} for {pretty_names_str}!"
                ),
                completed=total_files,
                total=total_files,
            )
            return responses


def _trend_table_for(
    id: ConnectableId,
    versions: Set[str],
    dialects_trend: dict[Dialect, _report.Report],
) -> Table:
    main_table = Table(show_lines=True)
    main_table.add_column(
        "Dialect",
        justify="center",
        vertical="middle",
    )
    main_table.add_column(
        f"Trend Data of {id} versions",
        justify="center",
    )

    def add_sub_table_row(
        sub_table: Table,
        version: str,
        unsuccessful: Unsuccessful,
    ):
        sub_table.add_row(
            version,
            str(len(unsuccessful.skipped)),
            str(len(unsuccessful.errored)),
            str(len(unsuccessful.failed)),
        )

    for dialect, report in dialects_trend.items():
        test = "tests" if report.total_tests != 1 else "test"
        sub_table = Table(
            "Version",
            "Skips",
            "Errors",
            "Failures",
            caption=f"{report.total_tests} {test} ran",
            show_edge=False,
        )

        if versions:
            for version, unsuccessful in report.latest_to_oldest():
                add_sub_table_row(sub_table, version, unsuccessful)
        else:
            implementation = report.implementations[id]
            version = implementation.version or "latest"
            unsuccessful = report.unsuccessful(id)
            add_sub_table_row(sub_table, version, unsuccessful)

        main_table.add_row(dialect.pretty_name, sub_table)

    return main_table


def _trend_table_in_markdown_for(
    id: ConnectableId,
    versions: Set[str],
    dialects_trend: dict[Dialect, _report.Report],
) -> str:
    rows_data: list[list[str]] = []

    inner_table_columns = [
        "Version",
        "Skips",
        "Errors",
        "Failures",
    ]

    def create_inner_table_row(
        version: str,
        unsuccessful: Unsuccessful,
    ):
        return [
            version,
            str(len(unsuccessful.skipped)),
            str(len(unsuccessful.errored)),
            str(len(unsuccessful.failed)),
        ]

    for dialect, report in dialects_trend.items():
        test = "tests" if report.total_tests != 1 else "test"
        inner_table_rows: list[list[str]] = []
        if versions:
            for version, unsuccessful in report.latest_to_oldest():
                inner_table_rows.append(
                    create_inner_table_row(version, unsuccessful),
                )
        else:
            implementation = report.implementations[id]
            version = implementation.version or "latest"
            unsuccessful = report.unsuccessful(id)
            inner_table_rows.append(
                create_inner_table_row(version, unsuccessful),
            )

        inner_markdown_table = convert_table_to_markdown(
            inner_table_columns,
            inner_table_rows,
        )
        row_data = [
            dialect.pretty_name,
            inner_markdown_table,
            f"**{report.total_tests} {test} ran**",
        ]
        rows_data.append(row_data)

    return f"## Trend Data of {id} versions:\n\n" + "\n\n".join(
        [
            f"### Dialect: {row_data[0]}\n{row_data[1]}\n\n{row_data[2]}"
            for row_data in rows_data
        ],
    )


class _VersionedReportsTar(click.File):
    """
    Select a tar containing previously produced versioned Bowtie reports.
    """

    name = "versioned_reports_tar"
    mode = "rb"

    def convert(  # type: ignore[reportIncompatibleMethodOverride]
        self,
        value: str | PathLike[str] | IO[Any],
        param: click.Parameter | None,
        ctx: click.Context,
    ) -> tuple[frozenset[str], Iterable[tuple[str, Dialect, _report.Report]]]:
        input = super().convert(value, param, ctx)

        id = cast(
            _connectables.Connectable,
            ctx.params.get("connectable"),
        ).to_terse()

        try:
            with tarfile.open(fileobj=input) as tar:
                members = tar.getmembers()

                if not any(
                    member.name.startswith(f"./{id}/") for member in members
                ):
                    STDERR.print(
                        f"Couldn't find a '{id}' directory in {input.name}.",
                    )
                    ctx.exit(EX.DATAERR)

                versions: frozenset[str] = frozenset()
                try:
                    versions_content = tar.extractfile(
                        f"./{id}/matrix-versions.json",
                    )
                except KeyError:
                    STDERR.print(
                        "No versions detected (couldn't find a "
                        "'matrix-versions.json' file in ",
                        f"{id} directory of {input.name}.)",
                    )
                    ctx.exit(EX.DATAERR)
                else:
                    if versions_content:
                        versions = frozenset(json.load(versions_content))

                versions_dirs = [
                    member.name
                    for member in members
                    if member.isdir()
                    and (
                        member.name.removeprefix(f"./{id}/").removeprefix("v")
                    )
                    in versions
                ]
                if not versions_dirs:
                    STDERR.print(
                        f"Couldn't find any versions sub-directories in {id} "
                        f"directory of {input.name} as per listed by its "
                        "matrix-versions.json file.",
                    )
                    ctx.exit(EX.DATAERR)

                progress = Progress(
                    TextColumn("[bold blue]{task.description}"),
                    SpinnerColumn(finished_text=""),
                    BarColumn(bar_width=None),
                    TextColumn(
                        "[progress.percentage]{task.percentage:>3.0f}%",
                    ),
                    "•",
                    MofNCompleteColumn(),
                    "•",
                    TimeElapsedColumn(),
                    console=console.Console(),
                    transient=True,
                )
                dialects = (
                    cast(
                        Iterable[Dialect] | None,
                        ctx.params.get("dialects"),
                    )
                    or Dialect.known()
                )
                pretty_names_str = pretty_names_str_for(dialects)
                task = progress.add_task(
                    description=(
                        f"Preparing to parse all versioned reports of "
                        f"{id} for {pretty_names_str} found in {input.name}"
                    ),
                    total=len(versions) * len(list(dialects)),
                )

                actual_parsed_files = 0
                versioned_reports: Iterable[
                    tuple[str, Dialect, _report.Report]
                ] = []

                with progress:
                    for version in versions:
                        for dialect in dialects:
                            progress.update(
                                task,
                                description=(
                                    f"Parsing: "
                                    f"v{version}/{dialect.short_name}.json"
                                ),
                            )
                            try:
                                report_content = tar.extractfile(
                                    f"./{id}/v{version}/{dialect.short_name}.json",
                                )
                            except KeyError:
                                versioned_reports.append(
                                    (
                                        version,
                                        dialect,
                                        _report.Report.empty(dialect=dialect),
                                    ),
                                )
                            else:
                                if report_content:
                                    versioned_reports.append(
                                        (
                                            version,
                                            dialect,
                                            _report.Report.from_serialized(
                                                TextIOWrapper(
                                                    report_content,
                                                    encoding="utf-8",
                                                ),
                                            ),
                                        ),
                                    )
                                    actual_parsed_files += 1
                            progress.update(task, advance=1)
                    progress.update(
                        task,
                        description=(
                            f"Successfully parsed all versioned reports "
                            f"of {id} for {pretty_names_str} found in "
                            f"{input.name}"
                        ),
                        completed=actual_parsed_files,
                        total=actual_parsed_files,
                    )
                    return versions, versioned_reports
        except tarfile.TarError:
            STDERR.print(f"Failed to process {input.name}.")
            ctx.exit(EX.DATAERR)


@subcommand
@click.option(
    "--implementation",
    "-i",
    "connectable",
    required=True,
    multiple=False,
    default=None,
    type=_connectables.ClickParam(),
    metavar="IMPLEMENTATION",
    help=(
        "A connectable ID for a JSON Schema implementation supported "
        "by Bowtie. Run `bowtie filter-implementations` for the full "
        "list of supported implementations."
    ),
)
@click.option(
    "--dialect",
    "-D",
    "dialects",
    multiple=True,
    default=lambda: Dialect.known(),
    type=_Dialect(),
    metavar="URI_OR_NAME",
    help=(
        "A URI or shortname identifying the dialect of the test suite "
        "you wish to see the trend data for. Possible shortnames include: "
        f"{POSSIBLE_DIALECT_SHORTNAMES}. If none specified then defaults "
        "to all known dialects."
    ),
)
@click.argument(
    "versioned_reports_tar",
    default=None,
    type=_VersionedReportsTar(mode="rb"),
    required=False,
)
@format_option()
def trend(
    connectable: Connectable,
    dialects: Iterable[Dialect],
    versioned_reports_tar: (
        tuple[
            frozenset[str],
            Iterable[tuple[str, Dialect, _report.Report]],
        ]
        | None
    ),
    format: _F,
):
    """
    Show trend data of an implementation across its multiple versions.
    """
    id = connectable.to_terse()

    if not versioned_reports_tar:

        async def download_versions_and_parse_reports_for(
            id: ConnectableId,
            dialects: Iterable[Dialect],
        ):
            versions = await download_versions_of(id)
            downloaded_versioned_reports = (
                await download_and_parse_reports_for(
                    id,
                    versions,
                    dialects,
                )
            )
            return versions, downloaded_versioned_reports

        versions, versioned_reports = asyncio.run(
            download_versions_and_parse_reports_for(id, dialects),
        )
    else:
        versions, versioned_reports = versioned_reports_tar

    if all(report.is_empty for _, _, report in versioned_reports):
        click.echo(
            f"None of the versions of '{id}' "
            f"support {pretty_names_str_for(dialects)}.",
        )
        return

    if versions:
        combine_versioned_reports_for = (
            _report.Report.combine_versioned_reports_for
        )
        dialects_trend = {
            dialect: combined_versions_report
            for dialect in sorted(dialects, reverse=True)
            if (
                combined_versions_report := combine_versioned_reports_for(
                    [report for _, _, report in versioned_reports],
                    dialect,
                )
            )
            and not combined_versions_report.is_empty
        }
    else:
        dialects_trend = {
            dialect: latest_report
            for _, dialect, latest_report in versioned_reports
            if not latest_report.is_empty
            and latest_report.implementations.get(id)
        }

    match format:
        case "json":
            serializable: Iterable[
                tuple[
                    ConnectableId,
                    Iterable[tuple[str, Iterable[tuple[str, dict[str, int]]]]],
                ]
            ] = [
                (
                    id,
                    [
                        (
                            dialect.short_name,
                            [
                                (
                                    version or "latest",
                                    {
                                        "failed": len(unsuccessful.failed),
                                        "errored": len(unsuccessful.errored),
                                        "skipped": len(unsuccessful.skipped),
                                    },
                                )
                                for version, unsuccessful in (
                                    report.latest_to_oldest()
                                    if versions
                                    else [
                                        (
                                            report.implementations[id].version,
                                            report.unsuccessful(id),
                                        ),
                                    ]
                                )
                            ],
                        )
                        for dialect, report in dialects_trend.items()
                    ],
                ),
            ]
            click.echo(json.dumps(serializable, indent=2))
        case "pretty":
            console.Console().print(
                _trend_table_for(id, versions, dialects_trend),
            )
        case "markdown":
            console.Console().print(
                _trend_table_in_markdown_for(id, versions, dialects_trend),
            )


@implementation_subcommand()  # type: ignore[reportArgumentType]
@click.option(
    "-q",
    "--quiet",
    "echo",
    # I have no idea why Click makes this so hard, but no combination of:
    #     type, default, is_flag, flag_value, nargs, ...
    # makes this work without doing it manually with callback.
    callback=lambda _, __, v: click.echo if not v else lambda *_, **__: None,  # type: ignore[reportUnknownLambdaType]
    is_flag=True,
    help="Don't print any output, just exit with nonzero status on failure.",
)
@format_option()
async def smoke(start: Starter, format: _F, echo: Callable[..., None]) -> int:
    """
    Smoke test implementations for basic correctness against Bowtie's protocol.
    """
    results = [
        (implementation.id, implementation.info, await implementation.smoke())
        async for _, implementation in start()
    ]

    match results, format:
        case [(_, _, result)], "json":
            echo(json.dumps(result.serializable(), indent=2))
        case _, "json":
            output = {id: result.serializable() for id, _, result in results}
            echo(json.dumps(output, indent=2))
        case [(_, _, result)], "pretty":
            console.Console().print(result)
        case _, "pretty":
            out = console.Console()
            for _, _, each in results:
                out.print(each)
        case _, "markdown":
            for _, info, result in results:
                echo(f"# {info.name} ({info.language})\n")

                if result.success:
                    echo("Smoke test *succeeded!*")
                else:
                    echo("Smoke test **failed!**")

                epilog: Sequence[
                    tuple[Dialect, Sequence[tuple[TestCase, SeqResult]]]
                ] = []

                echo("\n## Dialects\n")
                for dialect, failures in result.for_each_dialect():
                    if failures:
                        epilog.append((dialect, failures))
                        suffix = " **(failed)**"
                    else:
                        suffix = ""
                    echo(f"* {dialect.pretty_name}{suffix}")

                if epilog:
                    echo("\n## Failures\n")

                    for dialect, failures in epilog:
                        output = dedent(
                            f"""
                            <details>
                            <summary>{dialect.pretty_name}</summary>
                            """,
                        )
                        echo(output)

                        for case, each in failures:
                            output = dedent(
                                f"""
                                ### Schema

                                ```json
                                {json.dumps(case.schema)}
                                ```

                                #### Instances

                                """,
                            )
                            echo(output)

                            # FIXME: This will be nicer if/when Unsuccessful
                            #        contains the unsuccessful results.
                            for i, test in enumerate(case.tests):
                                result = each.result_for(i)
                                if TestResult(valid=test.expected()) != result:  # type: ignore[reportArgumentType]
                                    echo(f"* `{test.instance}`")

                        echo("\n</details>")

    return 0 if all(result.success for _, _, result in results) else EX.DATAERR


@subcommand
@IMPLEMENTATION
@FILTER
@fail_fast
@SET_SCHEMA
@VALIDATE
@click.argument("input", type=_suite.ClickParam(), metavar="DIALECT")
def suite(
    input: tuple[Iterable[TestCase], Dialect, dict[str, Any]],
    filter: CaseTransform,
    **kwargs: Any,
):
    """
    Run the official JSON Schema test suite against any implementation.

    Supports a number of possible inputs:

        * file paths found on the local file system containing tests, e.g.:

            - ``{PATH}/tests/draft7`` to run the draft 7 version's tests out of a local checkout of the test suite

            - ``{PATH}/tests/draft7/foo.json`` to run just one file from a checkout

        * URLs to the test suite repository hosted on GitHub, e.g.:

            - ``https://github.com/json-schema-org/JSON-Schema-Test-Suite/blob/main/tests/draft7/``
              to run a version directly from any branch which exists in GitHub

            - ``https://github.com/json-schema-org/JSON-Schema-Test-Suite/blob/main/tests/draft7/foo.json``
              to run a single file directly from a branch which exists in GitHub

        * short name versions of the previous URLs (similar to those providable
          to `bowtie validate --dialect`, e.g.:

            - ``7``, to run the draft 7 tests directly from GitHub (as in the
              URL example above)

    """  # noqa: E501
    _cases, dialect, metadata = input
    cases = filter(_cases)
    task = _run(**kwargs, dialect=dialect, cases=cases, run_metadata=metadata)
    return asyncio.run(task)


async def _run(
    connectables: Iterable[Connectable],
    cases: Iterable[TestCase],
    dialect: Dialect,
    fail_fast: bool,
    maybe_set_schema: Callable[[Dialect], CaseTransform],
    max_fail: int | None = None,
    max_error: int | None = None,
    run_metadata: dict[str, Any] = {},
    reporter: _report.Reporter = _report.Reporter(),
    **kwargs: Any,
) -> int:
    exit_code = 0
    acknowledged: Mapping[ConnectableId, ImplementationInfo] = {}
    runners: list[DialectRunner] = []
    async with _start(
        connectables=connectables,
        reporter=reporter,
        **kwargs,
    ) as starting:
        for each in starting:
            try:
                _, implementation = await each
            except STARTUP_ERRORS as error:
                exit_code |= EX.CONFIG
                STDERR.print(error)
                continue

            try:
                runner = await implementation.start_speaking(dialect)
            except DialectError as error:
                exit_code |= EX.CONFIG
                STDERR.print(error)
            except UnsupportedDialect as error:
                STDERR.print(error)
            else:
                acknowledged[implementation.id] = implementation.info
                runners.append(runner)

        if not runners:
            STDERR.print(
                "[bold red]No implementations started successfully![/]",
            )
            return exit_code | EX.CONFIG

        reporter.ready(
            _report.RunMetadata(
                implementations=acknowledged,
                dialect=dialect,
                metadata=run_metadata,
            ),
        )

        count = 0
        should_stop = False
        unsucessful = Unsuccessful()

        # Just to complement bowtie perf (not for other purposes)
        time_taken_by_implementations = 0
        time_output_file = (
            Path(os.environ["TIME_OUTPUT_FILE"])
            if "TIME_OUTPUT_FILE" in os.environ
            else None
        )

        for count, case in enumerate(maybe_set_schema(dialect)(cases), 1):
            seq_case = SeqCase(seq=count, case=case)
            got_result = reporter.case_started(seq_case, dialect)

            responses = [seq_case.run(runner=runner) for runner in runners]
            st_time = perf_counter_ns()

            for each in asyncio.as_completed(responses):
                result = await each
                got_result(result=result)
                unsucessful += result.unsuccessful()
                time_taken_by_implementations += perf_counter_ns() - st_time
                if (max_fail and len(unsucessful.failed) >= max_fail) or (
                    max_error and len(unsucessful.errored) >= max_error
                ):
                    should_stop = True

            if should_stop:
                STDERR.print(
                    "[bold yellow]Stopping -- the maximum number of "
                    "unsuccessful tests was reached![/]",
                )
                break
        reporter.finished(did_fail_fast=should_stop)
        if count == 0:
            exit_code |= EX.NOINPUT
            STDERR.print("[bold red]No test cases ran.[/]")
        elif count > 1:  # XXX: Ugh, this should be removed when Reporter dies
            STDERR.print(f"Ran [green]{count}[/] test cases.")

        if time_output_file:
            with time_output_file.open("a") as file:
                file.write(f"{time_taken_by_implementations}\n")

    return exit_code


@asynccontextmanager
async def _start(
    connectables: Iterable[Connectable],
    **kwargs: Any,
):
    async def _connected(
        connectable: Connectable,
        **kwargs: Any,
    ):
        implementation = await stack.enter_async_context(
            connectable.connect(**kwargs),
        )
        return connectable.to_terse(), implementation

    async with AsyncExitStack() as stack:
        yield asyncio.as_completed(
            [_connected(each, **kwargs) for each in connectables],
        )


def _stderr_processor(file: TextIO) -> structlog.typing.Processor:
    def stderr_processor(
        logger: structlog.typing.BindableLogger,
        method_name: str,
        event_dict: structlog.typing.EventDict,
    ) -> structlog.typing.EventDict:
        for each in "stderr", "traceback":
            contents = event_dict.pop(each, None)
            if contents is not None:
                implementation = event_dict["logger_name"]
                title = f"[traceback.title]{implementation} [dim]({each})"
                console.Console(file=file, color_system="truecolor").print(
                    panel.Panel(
                        contents.rstrip("\n"),
                        title=title,
                        border_style="traceback.border",
                        expand=True,
                        padding=(1, 4),
                        highlight=True,
                    ),
                )

        return event_dict

    return stderr_processor


def _redirect_structlog(log_level: int, file: TextIO = sys.stderr):
    """
    Reconfigure structlog's defaults to go to the given location.
    """
    structlog.configure(
        processors=[
            structlog.processors.add_log_level,
            structlog.processors.StackInfoRenderer(),
            structlog.dev.set_exc_info,
            structlog.processors.TimeStamper(
                fmt="%Y-%m-%d %H:%M.%S",
                utc=False,
            ),
            _stderr_processor(file),
            structlog.dev.ConsoleRenderer(
                colors=getattr(file, "isatty", lambda: False)(),
            ),
        ],
        logger_factory=structlog.WriteLoggerFactory(file),
        wrapper_class=structlog.make_filtering_bound_logger(log_level),
    )



"""
Clunky pepperings-over of communication with containers.

Some of this is warts from aiodocker, but mixed in with fun
special-to-this-package logic occasionally.
"""

from __future__ import annotations

from collections import deque
from contextlib import AsyncExitStack, asynccontextmanager
from typing import TYPE_CHECKING
import asyncio
import json

from aiodocker import Docker
from attrs import field, frozen, mutable
import aiodocker.exceptions

from bowtie._core import InvalidResponse, Restarted
from bowtie.exceptions import (
    CannotConnect,
    GotStderr,
    NoSuchImplementation,
    StartupFailed,
)

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Awaitable, Callable
    from typing import Any

    import aiodocker.containers
    import aiodocker.stream  # noqa: TC004 ??? no it's not?

    from bowtie._commands import Message


IMAGE_REPOSITORY = "ghcr.io/bowtie-json-schema"


@frozen
class _ClosedStream(Exception):
    """
    The stream is closed, and we tried to send something on it.

    This exception should never bubble out of this module, it should be handled
    by the connection.
    """

    container: aiodocker.containers.DockerContainer


@mutable
class Stream:
    """
    Wrapper to make aiodocker's Stream more pleasant to use.
    """

    _stream: aiodocker.stream.Stream = field(repr=False, alias="stream")
    _container: aiodocker.containers.DockerContainer = field(
        repr=False,
        alias="container",
    )
    _read_timeout_sec: float | None = field(
        repr=False,
        alias="read_timeout_sec",
    )
    _buffer: deque[bytes] = field(factory=deque, alias="buffer")
    _last: bytes = b""

    @classmethod
    def attached_to(
        cls,
        container: aiodocker.containers.DockerContainer,
        **kwargs: Any,
    ) -> Stream:
        stream = container.attach(stdin=True, stdout=True, stderr=True)
        return cls(stream=stream, container=container, **kwargs)

    def _read_with_timeout(self) -> Awaitable[aiodocker.stream.Message | None]:
        read = self._stream.read_out()
        return asyncio.wait_for(read, timeout=self._read_timeout_sec)

    async def send(self, message: str) -> None:
        try:  # aiodocker doesn't appear to properly report stream closure
            await self._stream.write_in(message.encode())
        except aiodocker.exceptions.DockerError as err:
            raise _ClosedStream(self._container) from err
        except (AssertionError, AttributeError):
            raise _ClosedStream(self._container) from None

    async def receive(self) -> str:
        if self._buffer:
            return self._buffer.popleft().decode()

        while True:
            try:
                message = await self._read_with_timeout()
            except asyncio.exceptions.TimeoutError:
                info: dict[str, Any] = await self._container.show()  # type: ignore[reportUnknownMemberType]
                if not info["State"]["Running"]:
                    raise _ClosedStream(self._container)
                raise

            if message is not None:
                break
            info: dict[str, Any] = await self._container.show()  # type: ignore[reportUnknownMemberType]
            if not info["State"]["Running"]:
                raise _ClosedStream(self._container)

        if message.stream == 2:  # noqa: PLR2004
            data: list[bytes] = []

            while message.stream == 2:  # noqa: PLR2004
                data.append(message.data)
                try:
                    message = await self._read_with_timeout()
                except asyncio.exceptions.TimeoutError:
                    message = None
                if message is None:
                    raise GotStderr(b"".join(data))

        line: bytes
        rest: list[bytes]
        while True:
            line, *rest = message.data.split(b"\n")
            if rest:
                line, self._last = self._last + line, rest.pop()
                self._buffer.extend(rest)
                return line.decode()

            message = None
            while message is None:
                message = await self._read_with_timeout()
            self._last += line


@mutable
class Connection:
    """
    A connection with a restartable container over stdio via request/responses.

    Requests and responses are JSON-serializable messages, with serialization
    handled here.
    """

    _new_stream: Callable[[], Awaitable[Stream]] = field(
        repr=False,
        alias="new_stream",
    )

    # Maybe second versions of these will be useful also at the Implementation
    # level again, to control for non-protocol-related flakiness or slowness
    _restarts: int = field(default=10, repr=False, alias="restarts")

    #: A per-request number of retries, before giving up
    _retry: int = field(default=3, repr=False)

    _connected_to: Stream | None = None

    @property
    async def _stream(self) -> Stream:
        if self._connected_to is None:
            self._connected_to = await self._new_stream()
        return self._connected_to

    async def request(self, message: Message) -> Message | None:
        request = f"{json.dumps(message)}\n"

        try:
            await (await self._stream).send(request)
        except _ClosedStream:
            self._restarts -= 1
            self._connected_to = None
            raise Restarted() from None

        for _ in range(self._retry):
            try:
                response = await (await self._stream).receive()
            except asyncio.exceptions.TimeoutError:
                continue
            except _ClosedStream as err:
                stderr: list[str] = await err.container.log(stderr=True)  # type: ignore[reportUnknownVariableType]
                if stderr:
                    raise GotStderr("".join(stderr).encode())
                return

            try:
                return json.loads(response)
            except json.JSONDecodeError as err:
                raise InvalidResponse(contents=response) from err


def _float_or_none(value: str | float | None) -> float | None:
    """
    Coerce 0 to None, otherwise return a float.
    """
    if value is None:
        return value
    value = float(value)
    if value:
        return value
    return None


@frozen(kw_only=True)
class ConnectableImage:

    _id: str = field(
        converter=lambda value: (
            value if "/" in value else f"{IMAGE_REPOSITORY}/{value}"
        ),
        alias="id",
    )

    #: An explicit timeout to wait for each implementation to respond
    #: to *each* instance being validated. Set this to 0 if you wish
    #: to wait forever, though note that this means you may end up waiting
    #: ... forever!
    _read_timeout_sec: float | None = field(
        default=2.0,
        converter=_float_or_none,
        repr=False,
        alias="read_timeout_sec",
    )

    kind = "image"

    @asynccontextmanager
    async def connect(self) -> AsyncIterator[Connection]:
        async with AsyncExitStack() as stack:
            docker = await stack.enter_async_context(Docker())
            create = start_container_maybe_pull

            async def new_stream():
                nonlocal create

                try:
                    container = await create(
                        docker=docker,
                        image_name=self._id,
                    )
                except aiodocker.exceptions.DockerError as err:
                    if err.status != 900:  # noqa: PLR2004
                        raise
                    raise CannotConnect(
                        kind=self.kind,
                        id=self._id,
                        hint=(
                            "Can't connect to your container runtime "
                            "(e.g. podman or docker). "
                            "Ensure you have one installed, that you have "
                            "set the DOCKER_HOST environment variable if "
                            "needed, and that containers successfully start "
                            "if you directly run one outside of Bowtie."
                        ),
                    ) from err

                stack.push_async_callback(container.delete, force=True)  # type: ignore[reportUnknownMemberType]
                create = start_container

                try:
                    return Stream.attached_to(
                        container,
                        read_timeout_sec=self._read_timeout_sec,
                    )
                except GotStderr as error:
                    err = StartupFailed(
                        id=self._id,
                        stderr=error.stderr.decode(),
                    )
                    raise err from None
                except _ClosedStream:
                    raise StartupFailed(id=self._id) from None

            yield Connection(new_stream=new_stream)


async def start_container_maybe_pull(docker: Docker, image_name: str):
    # You would think we would use aiodocker's container.start() function
    # which essentially does the below. You would think wrong.
    # That function will pull the *entire* image repository if it ends up
    # pulling our harness image -- so here we reimplement it, but only
    # pull :latest when the image is missing.
    try:
        return await start_container(docker=docker, image_name=image_name)
    except aiodocker.exceptions.DockerError as err:
        if err.status != 404:  # noqa: PLR2004
            raise

        try:
            tag = image_name.partition(":")[2] or "latest"
            await docker.pull(from_image=image_name, tag=tag)
        except aiodocker.exceptions.DockerError as err:
            # This craziness can go wrong in various ways, none of them
            # machine parseable.

            status, data, *_ = err.args
            if data.get("cause") == "image not known":
                raise NoSuchImplementation(image_name) from err

            message = ghcr = data.get("message", "")

            if status == 500:  # noqa: PLR2004
                try:
                    # GitHub Registry saying an image doesn't exist as
                    # reported within GitHub Actions' version of Podman...
                    # This is some crazy string like:
                    #   Head "https://ghcr.io/v2/bowtie-json-schema/image-name/manifests/latest": denied  # noqa: E501
                    # with seemingly no other indication elsewhere and
                    # obviously no good way to detect this specific case
                    no_image = message.endswith('/latest": denied')
                except Exception:  # noqa: BLE001, S110
                    pass
                else:
                    if no_image:
                        raise NoSuchImplementation(image_name)

                try:
                    # GitHub Registry saying an image doesn't exist as
                    # reported locally via podman on macOS...

                    # message will be ... a JSON string !?! ...
                    error = json.loads(ghcr).get("message", "")
                except Exception:  # noqa: BLE001, S110
                    pass  # nonJSON / missing key
                else:
                    if "403 (forbidden)" in error.casefold():
                        raise NoSuchImplementation(image_name)

            raise StartupFailed(id=image_name, data=data) from err
        return await start_container(docker=docker, image_name=image_name)


async def start_container(docker: Docker, image_name: str):
    config = dict(
        Image=image_name,
        OpenStdin=True,
        HostConfig=dict(NetworkMode="none"),
    )
    container = await docker.containers.create(config=config)
    await container.start()  # type: ignore[reportUnknownMemberType]
    return container


@frozen(kw_only=True)
class ConnectableContainer:

    _id: str = field(alias="id")

    #: An explicit timeout to wait for each implementation to respond
    #: to *each* instance being validated. Set this to 0 if you wish
    #: to wait forever, though note that this means you may end up waiting
    #: ... forever!
    _read_timeout_sec: float | None = field(
        default=2.0,
        converter=_float_or_none,
        repr=False,
        alias="read_timeout_sec",
    )

    kind = "container"

    @asynccontextmanager
    async def connect(self) -> AsyncIterator[Connection]:
        async with Docker() as docker:
            try:
                container = await docker.containers.get(self._id)  # type: ignore[reportUnknownMemberType]
            except aiodocker.exceptions.DockerError as err:
                raise CannotConnect(kind=self.kind, id=self._id) from err

            async def new_stream():
                return Stream.attached_to(
                    container,
                    read_timeout_sec=self._read_timeout_sec,
                )

            yield Connection(new_stream=new_stream)



from itertools import cycle
from pathlib import Path

from url.url import URL

from bowtie._benchmarks import BenchmarkGroup
from bowtie._core import Dialect


def get_benchmark():
    return BenchmarkGroup.from_dict(
        data=dict(
            name="nested_schemas",
            benchmark_type="default",
            dialects_supported=[
                Dialect.from_str(
                    "https://json-schema.org/draft/2020-12/schema",
                ),
            ],
            description=(
                "Validating highly nested schemas shouldn't "
                "cause exponential time blowups."
            ),
            schema={
                "$id": "https://example.com/draft/2020-12/schema/strict",
                "$schema": "https://json-schema.org/draft/2020-12/schema",
                "$vocabulary": {
                    "https://json-schema.org/draft/2020-12/vocab/core": True,
                    "https://json-schema.org/draft/2020-12/vocab/applicator": True,
                    "https://json-schema.org/draft/2020-12/vocab/unevaluated": True,
                    "https://json-schema.org/draft/2020-12/vocab/validation": True,
                    "https://json-schema.org/draft/2020-12/vocab/meta-data": True,
                    "https://json-schema.org/draft/2020-12/vocab/format-annotation": True,
                    "https://json-schema.org/draft/2020-12/vocab/content": True,
                },
                "$dynamicAnchor": "meta",
                "$ref": "https://json-schema.org/draft/2020-12/schema",
                "unevaluatedProperties": False,
            },
            tests=[
                dict(
                    description=f"No of Levels {levels}",
                    instance=nested_object(levels=levels),
                )
                for levels in range(1, 11, 3)
            ],
        ),
        uri=URL.parse(Path(__file__).absolute().as_uri()),
    )


def nested_object(levels: int):
    """
    Produce a schema which validates deeply nested objects and arrays.
    """
    names = cycle(["foo", "bar", "baz", "quux", "spam", "eggs"])
    schema = {
        "type": "object",
        "properties": {"ham": {"type": "string"}},
    }
    for _, name in zip(range(levels - 1), names):
        schema = {"type": "object", "properties": {name: schema}}
    return schema



from pathlib import Path

from jsonschema_specifications import REGISTRY as SPECIFICATIONS
from url.url import URL

from bowtie._benchmarks import BenchmarkGroup
from bowtie._core import Dialect

DRAFT202012_DIALECT_URI = "https://json-schema.org/draft/2020-12/schema"


def get_benchmark():
    return BenchmarkGroup.from_dict(
        data=dict(
            name="Draft2020-12_MetaSchema",
            benchmark_type="default",
            dialects_supported=[
                Dialect.from_str(
                    "https://json-schema.org/draft/2020-12/schema",
                ),
            ],
            description=(
                "A benchmark for validation of the Draft2020-12 MetaSchema."
            ),
            schema=SPECIFICATIONS.contents(DRAFT202012_DIALECT_URI),
            tests=[
                dict(
                    description="Validating metaschema against metaschema",
                    instance=SPECIFICATIONS.contents(
                        DRAFT202012_DIALECT_URI,
                    ),
                ),
            ],
        ),
        uri=URL.parse(Path(__file__).absolute().as_uri()),
    )



"""
Smoke testing of implementations.

Represents a sort of "absolute minimum level of assumed correctness" which we
can use to decide whether a harness is correctly written, even though
implementations may still be buggy / not fully compliant.
"""

from __future__ import annotations

from functools import cached_property
from textwrap import dedent
from typing import TYPE_CHECKING, Any
import json

from attrs import evolve, field, frozen
from referencing.jsonschema import EMPTY_REGISTRY
from rpds import HashTrieMap

from bowtie._commands import TestResult
from bowtie._core import Example, Test, TestCase

if TYPE_CHECKING:
    from collections.abc import Sequence

    from rich.console import Console, ConsoleOptions, RenderResult

    from bowtie._commands import SeqResult
    from bowtie._core import ConnectableId, Dialect, Implementation

EXAMPLES = [
    Example(description="null", instance=None),
    Example(description="boolean", instance=True),
    Example(description="integer", instance=37),
    Example(description="number", instance=37.37),
    Example(description="string", instance="37"),
    Example(description="array", instance=[37]),
    Example(description="object", instance={"foo": 37}),
]


async def test(implementation: Implementation):
    """
    Smoke test an implementation for absolute basic correctness.

    All implementations are expected to pass this test, and failing it is used
    to indicate that a *harness* is misimplemented.

    So nothing terribly complex should be included here, as elsewhere Bowtie of
    course can report on overall *compliance* of an implementation in order to
    communicate how well it implements a specification.

    Right now this test consists of verifying that the implementation properly:

        * validates with a schema which should consider all instances valid
        * validates with a schema which should consider all instances invalid

    In addition, we check, but do not fail, an implementation which does not
    support *basic* referencing support, where we define this to mean that a
    schema containing a reference to a single absolute URI is followed.
    """
    dialects = DialectResults()

    for dialect in implementation.info.dialects:
        cases = [
            dialect.top_test_case(EXAMPLES),
            dialect.bottom_test_case(EXAMPLES),
        ]
        failures = await implementation.failing(dialect=dialect, cases=cases)
        dialects = dialects.with_result(dialect, failures)

    # no point checking $ref unless we have a working dialect
    last, ref = dialects.latest_successful, None
    if last is not None:
        # We'd use a tag URI, but the goal is to use literally the simplest
        # possible thing that should work.
        simple_uri = "http://bowtie.report/cli/smoke/ref-to-string"
        resource = last.specification().create_resource({"type": "string"})
        ref_check = TestCase(
            description="$ref / registry support",
            schema={"$ref": simple_uri},
            registry=EMPTY_REGISTRY.with_resource(
                uri=simple_uri,
                resource=resource,
            ),
            tests=[
                Test(description="string", instance="valid", valid=True),
                Test(description="non-string", instance=37, valid=False),
            ],
        ).with_explicit_dialect(last)
        got = await implementation.failing(dialect=last, cases=[ref_check])
        if got:
            (failure,) = got
            ref = last, *failure

    return Result(id=implementation.id, dialects=dialects, ref=ref)


@frozen
class Result:
    """
    The result of smoke testing an implementation.
    """

    id: ConnectableId
    _dialects: DialectResults = field(repr=False, alias="dialects")
    _ref: tuple[Dialect, TestCase, SeqResult] | None = field(
        repr=False,
        alias="ref",
    )

    def __rich_console__(
        self,
        console: Console,
        options: ConsoleOptions,
    ) -> RenderResult:
        from rich import box
        from rich.console import Group
        from rich.json import JSON
        from rich.panel import Panel
        from rich.table import Column, Table
        from rich.text import Text

        epilog = Table.grid(padding=2, pad_edge=True)

        failures = self._dialects.failures

        prefix = f"{self.id} [dim]seems to"
        if failures:
            title = f"{prefix} be [/][b red]broken!"

            for dialect, cases_and_results in failures:
                first = cases_and_results[0][0]
                subtable = Table(
                    Column("Schema", vertical="middle", no_wrap=True),
                    Column("Instances", vertical="middle"),
                    title=f"{dialect.pretty_name} [red]Failures[/]",
                    padding=2,
                    box=box.MINIMAL,
                    show_header=False,
                    caption=dedent(
                        rf"""
                        [blue]bowtie validate -i {self.id} <(printf '
                            {json.dumps(first.schema)}
                        ') <(printf '
                            {json.dumps(first.tests[0].instance)}
                        ')

                        [/]can be used to reproduce one of the failures.
                        """,
                    ),
                    caption_justify="left",
                )
                for case, seq_result in cases_and_results:
                    instances = Table(box=None, padding=1)
                    for i, test in enumerate(case.tests):
                        result = seq_result.result.result_for(i)
                        if result.errored:
                            message = "[b red]errored"
                        else:
                            word = "valid" if result.valid else "invalid"  # type: ignore[reportUnknownMemberType]
                            message = f"[red]incorrectly[/] {word}"

                        instances.add_row(test.syntax(), message)

                    subtable.add_row(case.syntax(dialect), instances)
                epilog.add_row(subtable)
        elif self._ref:
            title = f"{prefix} be [/][b yellow]partially broken!"

            lead = Text(
                (
                    "The implementation was sent a simple `$ref` to "
                    "resolve and did not follow it correctly. "
                    "Check to be sure the harness is properly handling "
                    "the `registry` property when sent in a test case.\n"
                    "If however the implementation does not support "
                    "reference resolution in any form, you might not "
                    "be able to address this warning."
                ),
            )

            ref_dialect, case, failure = self._ref
            reason = Table(
                Column("Schema", vertical="middle"),
                Column("Registry", vertical="middle"),
                Column("Instances", vertical="middle"),
                box=box.MINIMAL,
            )

            # FIXME: Via python-jsonschema/referencing#16
            registry = {k: v.contents for k, v in case.registry.items()}
            instances = Table.grid(padding=2)
            for i, test in enumerate(case.tests):
                result = failure.result_for(i)
                expected = TestResult(valid=test.expected())  # type: ignore[reportArgumentType]
                if expected != result:
                    instances.add_row(
                        JSON(json.dumps(test.instance)),
                        (
                            f"[red]{result.description}[/] but should be "
                            f"[green]{expected.description}"
                        ),
                    )

            reason.add_row(
                case.syntax(ref_dialect),
                JSON(json.dumps(registry)),
                instances,
            )

            panel = Panel(
                Group(lead, reason),
                title="[b yellow]Basic referencing support does not work.",
                padding=2,
            )
            epilog.add_row(panel)
        else:
            title = f"{prefix} [/][b green]work!"

        yield ""
        table = Table(
            Column("Label", no_wrap=True, justify="center"),
            Column("Content"),
            title=title,
            padding=1,
            box=None,
            show_header=False,
            pad_edge=True,
            min_width=79,
            title_justify="center",
        )
        table.add_row("Dialects\n[dim](Confirmed Working)", self._dialects)
        yield table
        yield epilog

    def for_each_dialect(self):
        return sorted(self._dialects, reverse=True)

    @cached_property
    def success(self):
        # We treat referencing failures as soft failures, since so many
        # implementations have issues :( Perhaps this will change.
        return not self._dialects.failures

    def serializable(self) -> dict[str, Any]:
        if self.success:
            return dict(
                success=True,
                dialects=[
                    each.short_name for each, _ in self.for_each_dialect()
                ],
                **{"registry": False} if self._ref else {},
            )
        return dict(
            success=False,
            dialects={  # FIXME: some standard representation of discrepancy
                dialect.short_name: [
                    dict(
                        schema=case.schema,
                        instances=[each.instance for each in case.tests],
                        expected=[
                            dict(valid=expect) for expect in each.expected
                        ],
                        **each.result.serializable(),
                    )
                    for case, each in failures
                ]
                for dialect, failures in self.for_each_dialect()
            },
        )


@frozen
class DialectResults:
    """
    An indication of whether each claimed supported dialect works.
    """

    _dialects: HashTrieMap[Dialect, Sequence[tuple[TestCase, SeqResult]]] = (
        HashTrieMap()
    )
    latest_successful: Dialect | None = None

    def __iter__(self):
        return iter(self._dialects.items())

    @cached_property
    def failures(self):
        return sorted(
            ((k, v) for k, v in self._dialects.items() if v),
            reverse=True,
        )

    def __rich__(self):
        from rich.table import Column, Table

        table = Table.grid(Column("Name", justify="right"))
        ordered = sorted(self._dialects.items(), reverse=True)
        for dialect, failures in ordered:
            style = "red" if failures else "green"
            table.add_row(f"{dialect.pretty_name}", style=style)
        return table

    def with_result(
        self,
        dialect: Dialect,
        results: Sequence[tuple[TestCase, SeqResult]],
    ):
        latest = self.latest_successful
        if not results and (latest is None or dialect > latest):
            latest = dialect
        dialects = self._dialects.insert(dialect, results)
        return evolve(self, dialects=dialects, latest_successful=latest)



from __future__ import annotations

from collections.abc import Iterable
from contextlib import asynccontextmanager, suppress
from datetime import date
from functools import cache
from importlib.resources import files
from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol, cast
from uuid import uuid4
import json
import os

from attrs import Factory, asdict, evolve, field, frozen, mutable
from referencing.jsonschema import EMPTY_REGISTRY, specification_with
from rpds import HashTrieMap, HashTrieSet
from url import URL
import httpx
import referencing_loaders

from bowtie import HOMEPAGE, ORG_NAME
from bowtie._commands import (
    START_V1,
    CaseErrored,
    Dialect as DialectCommand,
    SeqCase,
    SeqResult,
    StartedDialect,
)
from bowtie.exceptions import (
    DialectError,
    GotStderr,
    InvalidResponse,
    ProtocolError,
    StartupFailed,
    UnsupportedDialect,
)

if TYPE_CHECKING:
    from collections.abc import (
        AsyncIterator,
        Callable,
        Mapping,
        Sequence,
        Set,
    )
    from typing import Self

    from referencing import Specification
    from referencing.jsonschema import Schema, SchemaRegistry
    from rich.console import RenderableType

    from bowtie._commands import (
        AnyCaseResult,
        Command,
        Message,
        Run,
        Seq,
    )
    from bowtie._connectables import ConnectableId
    from bowtie._registry import ValidatorRegistry
    from bowtie._report import Reporter

ORG_API = URL.parse("https://api.github.com/orgs/") / ORG_NAME
CONTAINER_PACKAGES_API = ORG_API / "packages" / "container"


@frozen
class Dialect:
    """
    A dialect of JSON Schema.
    """

    pretty_name: str
    uri: URL = field(repr=False)
    short_name: str = field(repr=False)
    first_publication_date: date = field(repr=False)
    aliases: Set[str] = field(
        default=cast(frozenset[str], frozenset()),
        repr=False,
    )
    has_boolean_schemas: bool = field(default=True, repr=False)

    _top_schema: Schema | None = field(
        default=Factory(
            lambda self: {"$schema": str(self.uri)},
            takes_self=True,
        ),
        hash=False,
        repr=False,
        alias="top_schema",
    )
    _bottom_schema: Schema | None = field(
        default=Factory(
            lambda self: (
                None
                if self._top_schema is None
                else {"$schema": str(self.uri), "not": self.top_schema}
            ),
            takes_self=True,
        ),
        hash=False,
        repr=False,
        alias="bottom_schema",
    )

    def __lt__(self, other: Any):
        if other.__class__ is not Dialect:
            return NotImplemented
        return self.first_publication_date < other.first_publication_date

    @classmethod
    @cache
    def by_short_name(cls) -> HashTrieMap[str, Dialect]:
        return HashTrieMap((each.short_name, each) for each in cls.known())

    @classmethod
    @cache
    def by_alias(cls) -> HashTrieMap[str, Dialect]:
        return cls.by_short_name().update(
            (alias, dialect)
            for dialect in cls.known()
            for alias in dialect.aliases
        )

    @classmethod
    @cache
    def by_uri(cls) -> HashTrieMap[URL, Dialect]:
        return HashTrieMap((each.uri, each) for each in cls.known())

    @classmethod
    @cache
    def known(cls) -> Iterable[Dialect]:
        data = files("bowtie") / "data"
        if not data.is_dir():
            data = Path(__file__).parent.parent / "data"

        return HashTrieSet(
            Dialect.from_dict(**each)
            for each in json.loads(data.joinpath("dialects.json").read_text())
        )

    @classmethod
    @cache
    def latest(cls):
        """
        The latest dialect known to Bowtie.
        """
        return max(cls.known())

    @classmethod
    def from_dict(
        cls,
        firstPublicationDate: str,
        prettyName: str,
        shortName: str,
        uri: str,
        aliases: Iterable[str] = (),
        hasBooleanSchemas: bool = True,
        **kwargs: Any,
    ) -> Self:

        for each in "top", "bottom":
            if each in kwargs:
                kwargs[f"{each}_schema"] = kwargs.pop(each)

        del kwargs["$schema"]
        return cls(
            uri=URL.parse(uri),
            pretty_name=prettyName,
            short_name=shortName,
            first_publication_date=date.fromisoformat(firstPublicationDate),
            aliases=frozenset(aliases),
            has_boolean_schemas=hasBooleanSchemas,
            **kwargs,
        )

    @classmethod
    def from_str(cls, uri: str):
        url = URL.parse(uri)
        by_uri = cls.by_uri()

        if url.fragment is None:
            dialect = by_uri.get(url.with_fragment(""))
            if dialect is not None:
                return dialect

        return by_uri[url]

    async def latest_report(self):
        url = HOMEPAGE / f"{self.short_name}.json"
        async with httpx.AsyncClient(timeout=10) as client:
            return await client.get(str(url))

    def serializable(self):
        return str(self.uri)

    def specification(self, **kwargs: Any) -> Specification[Schema]:
        return specification_with(str(self.uri), **kwargs)

    @property
    def top_schema(self):
        if self._top_schema is None:
            raise ValueError(f"{self} has no top schema.")
        return self._top_schema

    @property
    def bottom_schema(self):
        if self._bottom_schema is None:
            raise ValueError(f"{self} has no bottom schema.")
        return self._bottom_schema

    def top(self):
        """
        Create a validator in this dialect which allows all instances.
        """
        from bowtie._direct_connectable import Direct

        validators = Direct.from_id("python-jsonschema").registry()
        return validators.for_schema(self.top_schema)

    def bottom(self):
        """
        Create a validator in this dialect which does not allow any instances.
        """
        from bowtie._direct_connectable import Direct

        validators = Direct.from_id("python-jsonschema").registry()
        return validators.for_schema(self.bottom_schema)

    def top_test_case(
        self,
        examples: Iterable[Example],
        description: str = "top allows everything",
    ):
        return TestCase(
            description=description,
            schema=self.top_schema,
            tests=[example.expect(True) for example in examples],
        )

    def bottom_test_case(
        self,
        examples: Iterable[Example],
        description: str = "bottom allows nothing",
    ):
        return TestCase(
            description=description,
            schema=self.bottom_schema,
            tests=[example.expect(False) for example in examples],
        )


class Restarted(Exception):
    """
    A connection was restarted, so we may need to replay some messages.
    """


@frozen
class Link:
    description: str
    url: URL

    @classmethod
    def from_dict(cls, description: str, url: str):
        return cls(description=description, url=URL.parse(url))

    def serializable(self):
        return dict(description=self.description, url=str(self.url))


@frozen(order=True)
class ImplementationInfo:
    # FIXME: Combine with / separate out more from `Implementation`

    name: str
    language: str
    homepage: URL
    issues: URL
    source: URL
    dialects: frozenset[Dialect]

    version: str | None = None
    language_version: str | None = None
    os: str | None = None
    os_version: str | None = None
    documentation: URL | None = None
    links: Iterable[Link] = ()

    @classmethod
    def from_dict(
        cls,
        homepage: str,
        issues: str,
        source: str,
        dialects: list[str],
        links: Iterable[dict[str, Any]] = (),
        **kwargs: Any,
    ):
        return cls(
            homepage=URL.parse(homepage),
            issues=URL.parse(issues),
            source=URL.parse(source),
            dialects=frozenset(Dialect.from_str(each) for each in dialects),
            links=[Link.from_dict(**each) for each in links],
            **kwargs,
        )

    @property
    def id(self):
        """
        A unique identifier we use to refer to the implementation.

        This ID is *independent* of the connectable (i.e. how we speak to the
        implementation) but should be unique across disparate implementations.
        """
        return f"{self.language}-{self.name}"

    def serializable(self):
        return {
            **{
                k: (
                    str(v)
                    if k
                    in {
                        "homepage",
                        "issues",
                        "source",
                        "documentation",
                    }
                    else v
                )
                for k, v in asdict(self, recurse=False).items()
                if v
            },
            "dialects": sorted(
                (str(dialect.uri) for dialect in self.dialects),
                reverse=True,
            ),
            "links": [link.serializable() for link in self.links],
        }


class Connection(Protocol):
    """
    A connection to a specific JSON Schema implementation.

    Concrete implementations of this protocol will decide what means of
    communication are used -- e.g. sockets, files, in-memory -- as
    well as how the JSON Schema implementation is running -- in a separate
    process, in-memory, et cetera.
    """

    async def request(self, message: Message) -> Message | None:
        """
        Send a request to the harness.
        """
        ...


@frozen
class HarnessClient:
    """
    A client which speaks to a specific running implementation harness.
    """

    _connection: Connection = field(alias="connection")

    _registry: ValidatorRegistry[Any] = field(alias="registry")

    # FIXME: Remove this somehow by making the state machine even more explicit
    #: A sequence of commands to replay if we end up restarting the connection.
    _if_replaying: Sequence[Command[Any]] = ()

    async def _get_back_up_to_date(self):
        for each in self._if_replaying:
            await self.request(each)  # TODO: response assert?

    async def transition[R](self, cmd: Command[R]) -> tuple[Self, R | None]:
        response = await self.request(cmd)
        harness = evolve(self, if_replaying=[*self._if_replaying, cmd])
        return harness, response

    async def request[R](self, cmd: Command[R]) -> R | None:
        """
        Send a given command to the implementation and return its response.
        """
        request = cmd.to_request(registry=self._registry)
        try:
            response = await self._connection.request(request)
        except Restarted:
            await self._get_back_up_to_date()
            # FIXME: Probably handle infinitely restarting harnesses
            response = await self._connection.request(request)
        if response is not None:
            return cmd.from_response(response, registry=self._registry)


@frozen
class DialectRunner:
    """
    A running implementation which is speaking a specific dialect.
    """

    dialect: Dialect
    implementation: ConnectableId
    _harness: HarnessClient = field(repr=False, alias="harness")

    schema_without_dialect: Callable[[Any], None] = field(
        repr=False,
        alias="schema_without_dialect",
    )

    @classmethod
    async def for_dialect(
        cls,
        dialect: Dialect,
        implementation: ConnectableId,
        harness: HarnessClient,
        reporter: Reporter,
    ):
        new_harness: HarnessClient
        response: StartedDialect
        new_harness, response = await harness.transition(
            DialectCommand(dialect=str(dialect.uri)),  # type: ignore[reportArgumentType]
        )

        if response == StartedDialect.OK:

            def schema_without_dialect(_: Any):  # type: ignore[reportRedeclaration]
                pass

        else:

            def schema_without_dialect(schema: Any):
                reporter.schema_without_dialect(
                    schema=schema,
                    implementation=implementation,
                    dialect=dialect,
                )

        return cls(
            dialect=dialect,
            implementation=implementation,
            harness=new_harness,
            schema_without_dialect=schema_without_dialect,
        )

    async def validate(
        self,
        run: Run,
        expected: Sequence[bool | None],
    ) -> SeqResult:
        try:
            response: tuple[Seq, int, AnyCaseResult] | None = (
                await self._harness.request(run)  # type: ignore[reportArgumentType]
            )
            if response is None:
                result = CaseErrored.uncaught()
            else:
                seq, length, result = response
                if seq != run.seq:
                    result = CaseErrored.uncaught(
                        message="mismatched seq",
                        expected=run.seq,
                        got=seq,
                        response=result,
                    )
                elif length and length != len(expected):
                    result = CaseErrored.uncaught(
                        message="wrong number of responses",
                        expected=len(expected),
                        got=length,
                        response=result,
                    )
        except GotStderr as error:
            result = CaseErrored.uncaught(stderr=error.stderr.decode("utf-8"))
        except InvalidResponse as error:
            result = CaseErrored.uncaught(response=error.contents)
        return SeqResult(
            seq=run.seq,
            implementation=self.implementation,
            expected=expected,
            result=result,  # type: ignore[reportUnknownArgumentType]  # pyright seems confused
        )


@mutable
class Implementation:
    """
    A running implementation under test.
    """

    id: ConnectableId
    info: ImplementationInfo
    _harness: HarnessClient = field(repr=False, alias="harness")
    _reporter: Reporter = field(alias="reporter")

    @classmethod
    def known(cls) -> Set[ConnectableId]:
        data = files("bowtie") / "data"
        if data.is_dir():
            path = data / "known_implementations.json"
            known = json.loads(path.read_text())
        else:
            root = Path(__file__).parent.parent
            dir = root.joinpath("implementations").iterdir()
            known = (d.name for d in dir if not d.name.startswith("."))
        return frozenset(known)

    @classmethod
    @asynccontextmanager
    async def start(
        cls,
        id: ConnectableId,
        reporter: Reporter,
        **kwargs: Any,
    ) -> AsyncIterator[Self]:
        _harness = HarnessClient(**kwargs)

        try:
            harness, started = await _harness.transition(START_V1)  # type: ignore[reportArgumentType]
        except ProtocolError as err:
            raise StartupFailed(id=id) from err
        except GotStderr as err:
            raise StartupFailed(id=id, stderr=err.stderr.decode()) from err
        else:
            if started is None:
                raise StartupFailed(id=id)

        info = ImplementationInfo.from_dict(**started.implementation)  # type: ignore[reportUnknownArgumentType]

        yield cls(harness=harness, id=id, info=info, reporter=reporter)

    def supports(self, *dialects: Dialect) -> bool:
        """
        Does the implementation support (all of) the given dialect(s)?
        """
        return self.info.dialects.issuperset(dialects)

    async def get_versions(self) -> Iterable[str]:
        from github3.exceptions import (  # type: ignore[reportMissingTypeStubs]
            GitHubError,
        )
        from github3.models import (  # type: ignore[reportMissingTypeStubs]
            GitHubCore,
        )

        url = CONTAINER_PACKAGES_API / self.id / "versions"

        gh = github()
        pages: list[GitHubCore] = []
        with suppress(GitHubError):
            pages = gh._iter(count=-1, url=str(url), cls=GitHubCore)  # type: ignore[reportPrivateUsage]

        versions: Set[str] = (
            {self.info.version} if self.info.version else set()
        )
        for page in pages:
            try:
                tags = cast(
                    Iterable[str],
                    page.as_dict()["metadata"]["container"]["tags"],
                )
            except KeyError:
                continue
            else:
                versions.update([tag for tag in tags if "." in tag])

        return sorted(versions, key=sortable_version_key, reverse=True)

    async def validate(
        self,
        dialect: Dialect,
        cases: Iterable[TestCase],
    ) -> AsyncIterator[tuple[TestCase, SeqResult]]:
        """
        Run a collection of test cases under the given dialect.
        """
        runner = await self.start_speaking(dialect)

        for case in cases:
            seq_case = SeqCase(seq=uuid4().hex, case=case)
            yield case, await seq_case.run(runner=runner)

    async def failing(
        self,
        dialect: Dialect,
        cases: Iterable[TestCase],
    ) -> Sequence[tuple[TestCase, SeqResult]]:
        """
        Run the given cases and yield ones which the implementation fails.
        """
        results = self.validate(dialect=dialect, cases=cases)
        return [each async for each in results if each[1].unsuccessful()]

    async def start_speaking(self, dialect: Dialect) -> DialectRunner:
        if not self.supports(dialect):
            raise UnsupportedDialect(implementation=self, dialect=dialect)
        try:
            return await DialectRunner.for_dialect(
                implementation=self.id,
                dialect=dialect,
                harness=self._harness,
                reporter=self._reporter,
            )
        except GotStderr as error:
            # the implementation failed on the dialect request.
            # there's likely no reason to continue, so we throw an exception
            raise DialectError(
                implementation=self,
                dialect=dialect,
                stderr=error.stderr,
            ) from error

    def smoke(self):
        """
        Smoke test this implementation.
        """
        from bowtie import _smoke

        return _smoke.test(self)


@frozen
class Example:
    """
    A validation example where we don't have any particularly expected result.
    """

    description: str
    instance: Any
    comment: str | None = None

    def expected(self) -> None:
        """
        We have no expected result.
        """
        return None

    def expect(self, valid: bool) -> Test:
        """
        Decide we really do expect some specific result.
        """
        return Test(**asdict(self), valid=valid)

    def syntax(self) -> RenderableType:
        from pygments.lexers.data import (  # type: ignore[reportMissingTypeStubs]
            JsonLexer,
        )
        from rich.syntax import Syntax

        return Syntax(
            json.dumps(self.instance),
            lexer=JsonLexer(),
            background_color="default",
            word_wrap=True,
        )

    @classmethod
    def from_dict(
        cls,
        instance: Any = None,
        valid: bool | None = None,
        **data: Any,
    ) -> Example | Test:
        if valid is None:
            return cls(**data, instance=instance)
        return Test(**data, instance=instance, valid=valid)


@frozen
class Test:
    """
    An `Example` with a specific expected result.
    """

    description: str
    instance: Any
    valid: bool
    comment: str | None = None

    def expected(self) -> bool:
        """
        Expect our expected validity result.
        """
        return self.valid

    def syntax(self) -> RenderableType:
        from pygments.lexers.data import (  # type: ignore[reportMissingTypeStubs]
            JsonLexer,
        )
        from rich.syntax import Syntax

        return Syntax(
            json.dumps(self.instance),
            lexer=JsonLexer(),
            background_color="default",
            word_wrap=True,
        )


@frozen
class Group:
    description: str
    children: Sequence[Group] | Sequence[LeafGroup]
    comment: str | None = None
    registry: SchemaRegistry = EMPTY_REGISTRY


@frozen
class LeafGroup:
    description: str
    schema: Schema
    children: Sequence[Example | Test]
    comment: str | None = None
    registry: SchemaRegistry = EMPTY_REGISTRY


@frozen
class TestCase:
    description: str
    schema: Any
    tests: Sequence[Example | Test]
    comment: str | None = None
    registry: SchemaRegistry = EMPTY_REGISTRY

    @classmethod
    def from_dict(
        cls,
        dialect: Dialect,
        tests: Iterable[dict[str, Any]],
        registry: Mapping[str, Schema] = {},
        **kwargs: Any,
    ):
        populated = EMPTY_REGISTRY.with_contents(
            registry.items(),
            default_specification=dialect.specification(),
        )
        return cls(
            tests=[Example.from_dict(**test) for test in tests],
            registry=populated,
            **kwargs,
        )

    def with_explicit_dialect(self, dialect: Dialect):
        """
        Return a version of this test case with an explicit dialect ID.
        """
        schema = self.schema
        if not isinstance(self.schema, bool):
            schema = {"$schema": str(dialect.uri), **self.schema}
        return evolve(self, schema=schema)

    def serializable(self) -> Message:
        as_dict = asdict(
            self,
            filter=lambda k, v: k.name != "registry"
            and (k.name != "comment" or v is not None),
        )
        if self.registry:
            # FIXME: Via python-jsonschema/referencing#16
            as_dict["registry"] = {
                k: v.contents for k, v in self.registry.items()
            }
        return as_dict

    def syntax(self, dialect: Dialect) -> RenderableType:
        from jsonschema_lexer import JSONSchemaLexer
        from rich.syntax import Syntax

        return Syntax(
            json.dumps(self.schema, indent=2),
            lexer=JSONSchemaLexer(str(dialect.uri)),
            background_color="default",
            word_wrap=True,
        )

    def uniq(self) -> str:
        """
        An internally used unique identifier when we want unique cases.

        Really this is just the JSON-serialized, normalized case.

        But that can change.
        """
        return json.dumps(self.serializable(), sort_keys=True)

    def expected_results(self) -> Sequence[bool | None]:
        return [each.expected() for each in self.tests]

    def without_expected_results(self) -> Message:
        serializable = self.serializable()
        serializable["tests"] = [
            {
                k: v
                for k, v in test.items()
                if k != "valid" and (k != "comment" or v is not None)
            }
            for test in serializable.pop("tests")
        ]
        return serializable


@cache
def registry():
    resources = referencing_loaders.from_traversable(files("bowtie.schemas"))
    return EMPTY_REGISTRY.with_resources(resources).crawl()


def convert_table_to_markdown(
    columns: list[str],
    rows: list[list[str]],
):
    widths = [max(len(row[i]) for row in rows) for i in range(len(columns))]
    rows = [[elt.center(w) for elt, w in zip(line, widths)] for line in rows]

    header = "| " + " | ".join(columns) + " |"
    border_left = "|:"
    border_center = ":|:"
    border_right = ":|"

    separator = (
        border_left
        + border_center.join(["-" * w for w in widths])
        + border_right
    )

    # body of the table
    body = [""] * len(rows)  # empty string list that we fill after
    for idx, line in enumerate(rows):
        # for each line, change the body at the correct index
        body[idx] = "| " + " | ".join(line) + " |"
    body = "\n".join(body)

    return f"\n{header}\n{separator}\n{body}"


def sortable_version_key(version: str):
    """
    Generate a sortable key for version strings like "1.2.3".
    """
    parts = version.split(".")
    return [int(part) if part.isdigit() else part for part in parts]


def github():
    """
    Construct a GitHub client, optionally looking for a token.

    This extra behavior is just useful in GitHub actions workflows, and
    presumably if ``github3.py`` was more active would be default behavior.
    """
    from github3 import GitHub  # type: ignore[reportMissingTypeStubs]

    return GitHub(token=os.environ.get("GITHUB_TOKEN", ""))



from pathlib import Path
import random
import string

from url.url import URL

from bowtie._benchmarks import Benchmark, BenchmarkGroup
from bowtie._core import Dialect

max_num = 100000


def get_benchmark():
    name = "patternProperties"
    benchmark_type = "keyword"
    description = (
        "A benchmark for measuring performance of the "
        "implementation for the patternProperties keyword."
    )
    max_num_properties = 100000
    varying_parameter = "Object length"

    benchmarks = []
    num_properties = 1000
    while num_properties <= max_num_properties:
        property_value_pairs = [
            (_random_string(), random.randint(0, max_num))
            for _ in range(num_properties)
        ]

        invalid_at_first = [
            (_random_string(), "string"),
        ] + property_value_pairs[1:]
        invalid_at_middle = (
            property_value_pairs[: num_properties // 2]
            + [(_random_string(), "string")]
            + property_value_pairs[num_properties // 2 : -1]
        )
        invalid_at_last = property_value_pairs[1:] + [
            (_random_string(), "string"),
        ]
        valid = property_value_pairs

        tests = (
            [
                dict(
                    description="Invalid Property at first",
                    instance=_get_instance_object(invalid_at_first),
                ),
                dict(
                    description="Invalid Property at middle",
                    instance=_get_instance_object(invalid_at_middle),
                ),
                dict(
                    description="Invalid Property at last",
                    instance=_get_instance_object(invalid_at_last),
                ),
                dict(
                    description="Valid",
                    instance=_get_instance_object(valid),
                ),
            ]
            if num_properties == max_num_properties
            else [
                dict(
                    description="Valid",
                    instance=_get_instance_object(valid),
                ),
            ]
        )

        benchmarks.append(
            Benchmark.from_dict(
                name=f"No. of Properties - {num_properties}",
                description=(
                    "Validating the `patternProperties` keyword with "
                    f"no. of properties {num_properties}."
                ),
                schema=dict(
                    type="object",
                    patternProperties={
                        "^[a-z]+$": {"type": "integer"},
                    },
                ),
                tests=tests,
            ),
        )

        num_properties *= 10

    return BenchmarkGroup(
        name=name,
        benchmark_type=benchmark_type,
        description=description,
        dialects_supported=[
            Dialect.from_str("https://json-schema.org/draft/2020-12/schema"),
            Dialect.from_str("https://json-schema.org/draft/2019-09/schema"),
            Dialect.from_str("http://json-schema.org/draft-07/schema#"),
            Dialect.from_str("http://json-schema.org/draft-06/schema#"),
            Dialect.from_str("http://json-schema.org/draft-04/schema#"),
            Dialect.from_str("http://json-schema.org/draft-03/schema#"),
        ],
        benchmarks=benchmarks,
        uri=URL.parse(Path(__file__).absolute().as_uri()),
        varying_parameter=varying_parameter,
    )


def _get_instance_object(property_value_pairs):
    return dict(property_value_pairs)


def _random_string():
    string_length = 100
    return "".join(
        random.choice(string.ascii_lowercase) for _ in range(string_length)
    )



#!/usr/bin/env python3
from dataclasses import dataclass
from importlib import metadata
import io
import json
import platform
import sys
import traceback

import fastjsonschema


@dataclass
class Runner:
    _started: bool = False
    _stdout: io.TextIOWrapper = sys.stdout

    def run(self, stdin=sys.stdin):
        for line in stdin:
            each = json.loads(line)
            cmd = each.pop("cmd")
            response = getattr(self, f"cmd_{cmd}")(**each)
            self._stdout.write(f"{json.dumps(response)}\n")
            self._stdout.flush()

    def cmd_start(self, version):
        assert version == 1
        self._started = True
        return dict(
            version=1,
            implementation=dict(
                language="python",
                name="fastjsonschema",
                version=metadata.version("fastjsonschema"),
                homepage="https://horejsek.github.io/python-fastjsonschema/",
                documentation=(
                    "https://horejsek.github.io/python-fastjsonschema/"
                ),
                issues=(
                    "https://github.com/horejsek/python-fastjsonschema/issues"
                ),
                source="https://github.com/horejsek/python-fastjsonschema",
                dialects=[
                    "http://json-schema.org/draft-07/schema#",
                    "http://json-schema.org/draft-06/schema#",
                    "http://json-schema.org/draft-04/schema#",
                ],
                os=platform.system(),
                os_version=platform.release(),
                language_version=platform.python_version(),
            ),
        )

    def cmd_dialect(self, dialect):
        assert self._started, "Not started!"
        return dict(ok=False)

    def cmd_run(self, case, seq):
        assert self._started, "Not started!"
        schema = case["schema"]
        try:
            # The registry parameter isn't used here when building cases.
            # Unless I'm missing it, it doesn't seem there's a way to register
            # schemas with fastjsonschema. It seems to use RefResolver-like
            # objects, but they're not exposed to end-user APIs like compile.

            validate = fastjsonschema.compile(schema)

            results = []

            for test in case["tests"]:
                try:
                    validate(test["instance"])
                except fastjsonschema.JsonSchemaException:
                    results.append({"valid": False})
                else:
                    results.append({"valid": True})

            return dict(seq=seq, results=results)
        except:  # noqa: E722  fastjsonschema can raise e.g. IndentationError
            return dict(
                errored=True,
                seq=seq,
                context={"traceback": traceback.format_exc()},
            )

    def cmd_stop(self):
        assert self._started, "Not started!"
        sys.exit(0)


Runner().run()



from pathlib import Path
import uuid

from url.url import URL

from bowtie._benchmarks import Benchmark, BenchmarkGroup
from bowtie._core import Dialect


def get_benchmark():
    name = "maxContains"
    benchmark_type = "keyword"
    description = (
        "A benchmark for measuring performance of the implementation "
        "for the maxContains keyword."
    )
    max_array_length = 100000
    varying_parameter = "Array length"

    array_length = 1000
    benchmarks = []
    while array_length <= max_array_length:
        array = [uuid.uuid4().hex for _ in range(array_length)]

        all_at_first = [1, 1, 1] + array[:-3]
        all_at_middle = (
            array[2 : array_length // 2]
            + [1, 1, 1]
            + array[array_length // 2 : -1]
        )
        all_at_last = array[:-3] + [1, 1, 1]
        valid = array[:-1] + [1]

        tests = (
            [
                dict(description="All at First", instance=all_at_first),
                dict(description="All at Middle", instance=all_at_middle),
                dict(description="All at Last", instance=all_at_last),
                dict(description="Valid", instance=valid),
            ]
            if array_length == max_array_length
            else [
                dict(description="Valid", instance=valid),
            ]
        )

        benchmarks.append(
            Benchmark.from_dict(
                name=f"Array length - {array_length}",
                description=(
                    f"Validating the `maxContains` keyword over array of length {array_length}."
                ),
                schema={
                    "type": "array",
                    "contains": {
                        "type": "integer",
                    },
                    "maxContains": 2,
                },
                tests=tests,
            ),
        )
        array_length *= 10

    return BenchmarkGroup(
        name=name,
        benchmark_type=benchmark_type,
        description=description,
        dialects_supported=[
            Dialect.from_str("https://json-schema.org/draft/2020-12/schema"),
            Dialect.from_str("https://json-schema.org/draft/2019-09/schema"),
        ],
        benchmarks=benchmarks,
        uri=URL.parse(Path(__file__).absolute().as_uri()),
        varying_parameter=varying_parameter,
    )



from pathlib import Path

from url.url import URL

from bowtie._benchmarks import Benchmark, BenchmarkGroup
from bowtie._core import Dialect


def get_benchmark():
    return BenchmarkGroup(
        name="benchmark",
        description="benchmark",
        benchmarks=[
            Benchmark.from_dict(
                name="benchmark",
                schema={
                    "type": "object",
                },
                description="benchmark",
                tests=[
                    {
                        "description": "test",
                        "instance": {},
                    },
                    {
                        "description": "test",
                        "instance": {},
                    },
                ],
            ),
        ],
        uri=URL.parse(Path(__file__).absolute().as_uri()),
        benchmark_type="test",
        dialects_supported=list(Dialect.known()),
    )



from __future__ import annotations

from datetime import UTC, datetime
from typing import TYPE_CHECKING, TypedDict
import importlib.metadata
import json
import sys

from attrs import asdict, field, frozen
from attrs.filters import exclude
from rpds import HashTrieMap
from url import URL
import structlog.stdlib

from bowtie._commands import Seq, SeqCase, SeqResult, Unsuccessful
from bowtie._core import Dialect, TestCase, sortable_version_key
from bowtie._direct_connectable import Direct

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Mapping
    from typing import Any, Literal, Self, TextIO

    from bowtie._commands import AnyTestResult
    from bowtie._connectables import ConnectableId
    from bowtie._core import Example, ImplementationInfo, Test


class InvalidReport(Exception):
    """
    The report is invalid.
    """


class EmptyReport(InvalidReport):
    """
    The report was totally empty.
    """


class DuplicateCase(InvalidReport):
    """
    A `Seq` appeared twice in the report.
    """


class MissingFooter(InvalidReport):
    """
    A report is missing its footer.

    Even though that only tells us whether the report failed fast, it might
    mean there's actual data missing too.
    """


def writer(file: TextIO = sys.stdout) -> Callable[..., Any]:
    return lambda **result: file.write(f"{json.dumps(result)}\n")  # type: ignore[reportUnknownArgumentType]


@frozen
class Reporter:
    _write: Callable[..., Any] = field(default=writer(), alias="write")
    _log: structlog.stdlib.BoundLogger = field(
        factory=structlog.stdlib.get_logger,
    )

    def schema_without_dialect(
        self,
        implementation: str,
        dialect: Dialect,
        schema: Any,
    ):
        self._log.warn(
            (
                f"The schema {schema!r} does not indicate its dialect via "
                "the $schema keyword. "
                f"This implementation does not support external dialect "
                "configuration, so validation results may not properly "
                "take the current dialect into account."
            ),
            logger_name=implementation,
            dialect=dialect.pretty_name,
        )

    def ready(self, run_metadata: RunMetadata):
        self._log.debug("Will speak", dialect=run_metadata.dialect)
        self._write(**run_metadata.serializable())

    def finished(self, did_fail_fast: bool):
        self._write(did_fail_fast=did_fail_fast)

    def case_started(self, seq_case: SeqCase, dialect: Dialect):
        self._write(**seq_case.serializable())
        log = self._log.bind(
            case=seq_case.case.description,
            schema=seq_case.case.schema,
        )

        if not seq_case.matches_dialect(dialect):
            log.warning(
                "$schema keyword does not seem to match the expected dialect",
                expected=dialect,
            )

        def got_result(result: SeqResult):
            bound = log.bind(logger_name=result.implementation)
            serialized = result.log_and_be_serialized(log=bound)
            self._write(**serialized)

        return got_result


@frozen
class RunMetadata:
    dialect: Dialect
    implementations: Mapping[ConnectableId, ImplementationInfo] = field(
        repr=lambda value: (
            f"({len(value)} implementation{'s' if len(value) != 1 else ''})"
        ),
        alias="implementations",
    )
    bowtie_version: str = field(
        default=importlib.metadata.version("bowtie-json-schema"),
        eq=False,
        repr=False,
    )
    metadata: Mapping[str, Any] = field(factory=dict, repr=False)
    started: datetime = field(
        factory=lambda: datetime.now(UTC),
        eq=False,
        repr=False,
    )

    @classmethod
    def from_dict(
        cls,
        dialect: str,
        implementations: dict[str, dict[str, Any]],
        started: str | None = None,
        **kwargs: Any,
    ) -> RunMetadata:
        from bowtie._core import ImplementationInfo

        if started is not None:
            kwargs["started"] = datetime.fromisoformat(started)
        return cls(
            dialect=Dialect.from_str(dialect),
            implementations={
                id: ImplementationInfo.from_dict(**data)
                for id, data in implementations.items()
            },
            **kwargs,
        )

    def serializable(self):
        as_dict = asdict(
            self,
            filter=exclude("dialect"),
            recurse=False,
        )
        as_dict.update(
            dialect=self.dialect.serializable(),
            started=as_dict.pop("started").isoformat(),
            # FIXME: This transformation is to support the UI parsing
            implementations={
                id: implementation.serializable()
                for id, implementation in self.implementations.items()
            },
        )
        return as_dict


@frozen(eq=False)
class Report:
    r"""
    A Bowtie report, containing (amongst other things) results about cases run.

    In general, reports are assumed to be constructed by calling
    `Report.from_serialized` or an equivalent (i.e. by 'replaying' JSON output
    that was produced by Bowtie).

    When comparing reports (e.g. for equality), only the relative order of
    test cases is considered, not the exact `Seq`\ s used in the run.
    """

    _cases: HashTrieMap[Seq, TestCase] = field(alias="cases", repr=False)
    _results: HashTrieMap[
        ConnectableId,
        HashTrieMap[Seq, SeqResult],
    ] = field(
        repr=False,
        alias="results",
    )
    metadata: RunMetadata
    did_fail_fast: bool

    def __eq__(self, other: object):
        if type(other) is not Report:
            return NotImplemented

        if (  # short circuit for fewer/more cases or differing implementations
            len(self._cases) != len(other._cases)
            or self._results.keys() != other._results.keys()
        ):
            return False

        this, that = asdict(self, recurse=False), asdict(other, recurse=False)

        cases = [v for _, v in sorted(this.pop("_cases").items())]
        if cases != [v for _, v in sorted(that.pop("_cases").items())]:
            return False

        other_results = that.pop("_results")
        for name, results in this.pop("_results").items():
            ordered = [v for _, v in sorted(results.items())]
            if ordered != [v for _, v in sorted(other_results[name].items())]:
                return False
        return this == that

    @classmethod
    def from_input(cls, input: Iterable[Mapping[str, Any]]) -> Self:
        # TODO: Support some interface for enabling/disabling validation.
        validator = (
            Direct.from_id("python-jsonschema")
            .registry()
            .for_uri(
                URL.parse("tag:bowtie.report,2024:report"),
            )
        )

        iterator = iter(input)
        header = next(iterator, None)
        if header is None:
            raise EmptyReport()
        metadata = RunMetadata.from_dict(**validator.validated(header))

        results: HashTrieMap[
            ConnectableId,
            HashTrieMap[Seq, SeqResult],
        ] = HashTrieMap.fromkeys(  # type: ignore[reportUnknownMemberType]
            metadata.implementations,
            HashTrieMap(),
        )
        cases: HashTrieMap[Seq, TestCase] = HashTrieMap()

        for data in iterator:
            match validator.validated(data):
                case {"seq": seq, "case": case}:
                    if seq in cases:
                        raise DuplicateCase(seq)
                    case = TestCase.from_dict(dialect=metadata.dialect, **case)
                    cases = cases.insert(seq, case)
                    continue
                case {"did_fail_fast": did_fail_fast}:
                    return cls(
                        results=results,
                        cases=cases,
                        metadata=metadata,
                        did_fail_fast=did_fail_fast,
                    )
                case _:
                    result = SeqResult.from_dict(**data)

            current = results.get(result.implementation, HashTrieMap())
            results = results.insert(  # TODO: Complain if present
                result.implementation,
                current.insert(result.seq, result),
            )

        raise MissingFooter()

    @classmethod
    def from_serialized(cls, serialized: Iterable[str]) -> Self:
        return cls.from_input(json.loads(line) for line in serialized)

    @classmethod
    def empty(
        cls,
        implementations: Mapping[ConnectableId, ImplementationInfo] = {},
        **kwargs: Any,
    ):
        """
        'The' empty report.
        """
        return cls(
            cases=HashTrieMap(),
            results=HashTrieMap(),
            metadata=RunMetadata(implementations=implementations, **kwargs),
            did_fail_fast=False,
        )

    @classmethod
    def combine_versioned_reports_for(
        cls,
        versioned_reports: Iterable[Report],
        dialect: Dialect,
    ) -> Report:
        versioned_reports = [
            versioned_report
            for versioned_report in versioned_reports
            if versioned_report.metadata.dialect == dialect
            and not versioned_report.is_empty
        ]
        if not versioned_reports:
            return cls.empty(dialect=dialect)

        results: HashTrieMap[
            ConnectableId,
            HashTrieMap[Seq, SeqResult],
        ] = HashTrieMap()
        implementations: dict[ConnectableId, ImplementationInfo] = {}

        for versioned_report in versioned_reports:
            ((version_id, version_info),) = (
                versioned_report.metadata.implementations.items()
            )
            implementations[version_id] = version_info
            results = results.insert(
                version_id,
                versioned_report._results[version_id],
            )

        return cls(
            cases=versioned_reports[0]._cases,
            results=results,
            metadata=RunMetadata(
                implementations=implementations,
                dialect=dialect,
            ),
            did_fail_fast=False,
        )

    @property
    def implementations(self) -> Mapping[ConnectableId, ImplementationInfo]:
        return self.metadata.implementations

    @property
    def is_empty(self):
        return not self._cases

    @property
    def total_tests(self):
        return sum(len(case.tests) for case in self._cases.values())

    def compliance_by_implementation(self):
        """
        Return the fraction of passing tests for each reported implementation.
        """
        return {
            id: 1 - (unsuccessful.total / self.total_tests)
            for id, _, unsuccessful in self.worst_to_best()
        }

    def unsuccessful(self, implementation: ConnectableId) -> Unsuccessful:
        """
        A count of the unsuccessful tests for the given implementation.
        """
        results = self._results[implementation].values()
        return sum((each.unsuccessful() for each in results), Unsuccessful())

    def worst_to_best(self):
        """
        All implementations ordered by number of unsuccessful tests.

        Ties are then broken alphabetically.
        """
        unsuccessful = [
            (id, implementation, self.unsuccessful(id))
            for id, implementation in self.implementations.items()
        ]
        unsuccessful.sort(key=lambda each: (each[2].total, each[1].name))
        return unsuccessful

    def latest_to_oldest(self):
        """
        Versioned implementations sorted by their latest to oldest versions.
        """
        unsuccessful = [
            (implementation.version, self.unsuccessful(id))
            for id, implementation in self.implementations.items()
            if implementation.version is not None
        ]
        unsuccessful.sort(
            key=lambda version_compliance: (
                sortable_version_key(version_compliance[0])
            ),
            reverse=True,
        )
        return unsuccessful

    def cases_with_results(
        self,
    ) -> Iterable[
        tuple[
            TestCase,
            Iterable[tuple[Example | Test, Mapping[str, AnyTestResult]]],
        ]
    ]:
        for seq, case in sorted(self._cases.items()):
            test_results: list[
                tuple[Example | Test, Mapping[str, AnyTestResult]]
            ] = []
            for i, test in enumerate(case.tests):
                test_result = {
                    id: self._results[id][seq].result_for(i)
                    for id in self.implementations
                }
                test_results.append((test, test_result))
            yield case, test_results

    def compliance_badges(self) -> Iterable[tuple[ImplementationInfo, Badge]]:
        for id, implementation in self.implementations.items():
            unsuccessful = self.unsuccessful(id)
            passed = self.total_tests - unsuccessful.total
            percentage = int(100 * (passed / self.total_tests))
            r, g, b = 100 - percentage, percentage, 0
            yield implementation, Badge(
                schemaVersion=1,
                label=self.metadata.dialect.pretty_name,
                message=f"{percentage}% Passing",
                color=f"{r:02x}{g:02x}{b:02x}",
            )


class Badge(TypedDict):
    schemaVersion: Literal[1]
    label: str
    message: str
    color: str


def supported_version_badge(dialects: Iterable[Dialect]) -> Badge:
    message = ", ".join(
        each.pretty_name.removeprefix("Draft ")
        for each in sorted(dialects, reverse=True)
    )
    return Badge(
        schemaVersion=1,
        label="JSON Schema Versions",
        message=message,
        color="lightgreen",
    )



import subprocess
import sys


def test_help_is_not_truncated():
    """
    Click will truncate short help messages if they are too long.

    That should never happen.
    """

    result = subprocess.run(
        [sys.executable, "-m", "bowtie", "--help"],
        capture_output=True,
        check=True,
    )
    stdout = result.stdout.decode().strip()
    truncated = [
        line  # [1:]: ignore the Usage: line
        for line in stdout.splitlines()[1:]
        if "..." in line
    ]
    assert not truncated, stdout


def test_commands_are_sorted_into_bins():
    """
    We want rich-click to be used to help users figure out which commands are
    relevant to them.

    Ideally it (or we) should have an API to *ensure* every subcommand we
    create goes in a bin, but until then let's assert against the "catch-all"
    bin appearing.
    """

    result = subprocess.run(
        [sys.executable, "-m", "bowtie", "--help"],
        capture_output=True,
        check=True,
    )
    stdout = result.stdout.decode().strip()
    assert not any("─ Commands ─" in i for i in stdout.splitlines()), stdout



"""
Direct connectables do not really connect anywhere and just operate in-memory.
"""

from __future__ import annotations

from contextlib import nullcontext
from importlib import metadata
from typing import TYPE_CHECKING, Any, Never
import pkgutil
import platform

from attrs import asdict, field, frozen, mutable
from referencing.jsonschema import EMPTY_REGISTRY
from url import URL

from bowtie import DOCS, HOMEPAGE, REPO
from bowtie._commands import CaseResult, Started, StartedDialect, TestResult
from bowtie._core import Dialect, ImplementationInfo, registry
from bowtie._registry import Invalid, SchemaCompiler, ValidatorRegistry
from bowtie.exceptions import CannotConnect

if TYPE_CHECKING:
    from collections.abc import Callable
    from contextlib import AbstractAsyncContextManager

    from jsonschema import ValidationError
    from jsonschema.protocols import Validator
    from referencing.jsonschema import Schema, SchemaRegistry

    from bowtie._commands import Message
    from bowtie._connectables import ConnectableId


def not_yet_connected(schema: Schema, registry: SchemaRegistry):
    def _not_yet_connected(instance: Any):
        raise RuntimeError("Not yet connected!")

    return _not_yet_connected


@mutable
class Unconnection[E: Exception]:

    _info: ImplementationInfo = field(repr=lambda i: i.id, alias="info")
    compiler_for: Callable[[Dialect], SchemaCompiler[E]] = field(
        repr=False,
        alias="compiler_for",
    )
    _implicit_dialect_response: StartedDialect = field(
        default=StartedDialect.OK,
        repr=False,
        alias="implicit_dialect_response",
    )

    _current_dialect: Dialect = Dialect.latest()
    _compile: SchemaCompiler[E] = not_yet_connected

    async def request(self, message: Message) -> Message:
        """
        Unpack the request and call our implementation.

        Call us, we always respond (so we never return ``None``).
        """
        match message:  # FIXME: Should request take Commands?
            case {"cmd": "start", "version": 1}:
                started = Started(
                    implementation=self._info.serializable(),  # FIXME
                    version=1,
                )
                return asdict(started)
            case {"cmd": "dialect", "dialect": uri}:
                self._current_dialect = Dialect.from_str(uri)
                self._compile = self.compiler_for(self._current_dialect)
                return asdict(self._implicit_dialect_response)
            case {"cmd": "run", "seq": seq, "case": case}:
                schema = case["schema"]
                registry = EMPTY_REGISTRY.with_contents(
                    case.get("registry", {}).items(),
                    default_specification=self._current_dialect.specification(),
                )
                validate = self._compile(schema, registry)
                results = [
                    TestResult(valid=validate(test["instance"]) is None)
                    for test in case["tests"]
                ]
                return {  # FIXME: Bleh this is not SeqResult
                    "seq": seq,
                    **CaseResult(results=results).serializable(),
                }
            case {"cmd": "stop"}:
                return {}
            case _:
                raise RuntimeError(f"Unknown message: {message!r}")


@frozen
class DirectImplementation[E: Exception]:

    _compiler_for: Callable[[Dialect], SchemaCompiler[E]]
    _info: ImplementationInfo
    _implicit_dialect_response: StartedDialect

    def __init__(
        self,
        compiler_for: Callable[[Dialect], SchemaCompiler[E]],
        implicit_dialect_response: StartedDialect = StartedDialect.OK,
        **kwargs: Any,
    ):
        object.__setattr__(self, "_compiler_for", compiler_for)
        object.__setattr__(
            self,
            "_implicit_dialect_response",
            implicit_dialect_response,
        )

        name = kwargs.pop("name", compiler_for.__name__)
        if "version" not in kwargs:
            kwargs["version"] = metadata.version(name)
        info = ImplementationInfo(
            name=name,
            os=platform.system(),
            os_version=platform.release(),
            language_version=platform.python_version(),
            **kwargs,
        )
        object.__setattr__(self, "_info", info)

    def __call__(self):
        return Unconnection(
            compiler_for=self._compiler_for,
            info=self._info,
            implicit_dialect_response=self._implicit_dialect_response,
        )

    @classmethod
    def from_callable(
        cls,
        language: str = "python",
        **kwargs: Any,
    ):
        def decorator(fn: Callable[[Dialect], SchemaCompiler[E]]):
            return lambda: cls(compiler_for=fn, language=language, **kwargs)

        return decorator


@DirectImplementation.from_callable(
    language="python",
    homepage=URL.parse("https://python-jsonschema.readthedocs.io/"),
    documentation=URL.parse("https://python-jsonschema.readthedocs.io/"),
    issues=URL.parse("https://github.com/python-jsonschema/jsonschema/issues"),
    source=URL.parse("https://github.com/python-jsonschema/jsonschema"),
    dialects=frozenset(
        [
            Dialect.by_short_name()["draft2020-12"],
            Dialect.by_short_name()["draft2019-09"],
            Dialect.by_short_name()["draft7"],
            Dialect.by_short_name()["draft6"],
            Dialect.by_short_name()["draft4"],
            Dialect.by_short_name()["draft3"],
        ],
    ),
)
def jsonschema(dialect: Dialect) -> SchemaCompiler[ValidationError]:
    from jsonschema.validators import (
        validator_for,  # type: ignore[reportUnknownVariableType]
    )

    # FIXME: python-jsonschema/jsonschema#1011
    def to_group(error: ValidationError) -> ValidationError | Invalid[Any]:
        """
        Upconvert a validation error to an exception group.
        """
        if error.context:  # FIXME: Throws away all the other attributes...
            return Invalid(
                error.message,
                [to_group(each) for each in error.context],
            )
        elif error.cause:
            return Invalid(error.message, [error.cause])
        return error

    def compile(
        schema: Schema,
        registry: SchemaRegistry,
    ) -> Callable[[Any], Invalid[ValidationError] | None]:
        DialectValidator: type[Validator] = validator_for(  # type: ignore[reportUnknownVariableType]
            schema,
            default=validator_for({"$schema": str(dialect.uri)}),
        )
        validator: Validator = DialectValidator(schema, registry=registry)  # type: ignore[reportUnknownVariableType]

        def validate(instance: Any) -> Invalid[Any] | None:
            errors = validator.iter_errors(instance)  # type: ignore[reportUnknownMemberType]
            exceptions: list[Exception] = [to_group(each) for each in errors]  # type: ignore[reportUnknownArgumentType]
            if exceptions:
                return Invalid("Not valid.", exceptions)

        return validate

    return compile


# pyright (at least 1.1.372) seems unable to see through this type if we try to
# define it via DirectImplementation.from_callable. But I guess that's fine...
def null(dialect: Dialect) -> SchemaCompiler[Never]:
    return lambda _, __: lambda _: None


NULL = DirectImplementation(
    compiler_for=null,
    language="python",
    version=metadata.version("bowtie-json-schema"),
    homepage=HOMEPAGE,
    documentation=DOCS,
    issues=REPO / "issues",
    source=REPO,
    dialects=frozenset(Dialect.known()),
)


IMPLEMENTATIONS: dict[str, Callable[..., Callable[[], Unconnection[Any]]]] = {
    "python-jsonschema": jsonschema,  # type: ignore[reportAssignmentType]  # pyright seems confused.
}


@frozen(kw_only=True)
class Direct[E: Exception]:
    """
    A direct connectable connects by simply importing some Python object.

    This is generally going to be suitable only to implementations written in
    Python or to a language where a wrapper (PyO3, CFFI, ...) can expose
    libraries written in the language directly to Python.

    This object represents only the lazy import.
    Connecting will lookup and call the object from the identified location.
    The return value should then be an object which makes use of a specific
    target implementation.

    Note that unlike when running implementations via containers, here we do
    *not* (cannot) restrict network access. Implementations making use of this
    functionaly should still ensure no networking takes place.
    """

    _wraps: Callable[[], Unconnection[E]] = field(alias="wraps")

    kind = "direct"

    @classmethod
    def from_id(cls, id: ConnectableId, **params: Any) -> Direct[Any]:
        if id == "null":
            return cls.null()

        wrapper = IMPLEMENTATIONS.get(id)
        if wrapper is not None:
            return cls(wraps=wrapper(**params))

        try:
            wrapper = pkgutil.resolve_name(id)
        except (ModuleNotFoundError, ValueError) as err:
            raise CannotConnect(kind=cls.kind, id=id) from err
        else:
            # pkgutil.resolve_name supports foo.bar.baz in addition to
            # foo.bar:baz. The latter has minor performance benefits, and in
            # addition, in our case it's easy to accidentally forget
            # *parameters* that foo.bar.baz takes, so here we disallow any
            # import string which does not use the foo.bar:baz syntax even
            # though it has succeeded.
            if ":" not in id:
                corrected = ":".join(id.rsplit(".", 1))
                raise CannotConnect(
                    kind=cls.kind,
                    id=id,
                    hint=(
                        f"{id!r} is not a known 'named' direct connectable "
                        "and instead appears to describe an object to "
                        "import, but it is not well formed. "
                        f"You may mean to use {corrected!r}."
                    ),
                )
        return cls(wraps=wrapper(**params))

    @classmethod
    def null(cls):
        """
        The null direct connectable considers all instances always valid.

        It can be useful e.g. for simply testing whether something is valid
        JSON, or for disabling validation where it otherwise would happen.
        """
        return cls(wraps=NULL)  # type: ignore[reportArgumentType]

    def connect(
        self,
        **kwargs: Any,
    ) -> AbstractAsyncContextManager[Unconnection[E]]:
        return nullcontext(self._wraps())

    def registry(self, **kwargs: Any) -> ValidatorRegistry[E]:
        if "registry" not in kwargs:
            kwargs["registry"] = registry()
        return ValidatorRegistry(
            compile=self._wraps().compiler_for(Dialect.latest()),  # XXX
            **kwargs,
        )



from pathlib import Path

from url.url import URL

from bowtie._benchmarks import BenchmarkGroup
from bowtie._core import Dialect


def get_benchmark():
    num_useless_keywords = 300000

    return BenchmarkGroup.from_dict(
        data=dict(
            name="useless_keywords",
            benchmark_type="default",
            dialects_supported=[
                Dialect.from_str(
                    "https://json-schema.org/draft/2020-12/schema",
                ),
                Dialect.from_str(
                    "https://json-schema.org/draft/2019-09/schema",
                ),
                Dialect.from_str("http://json-schema.org/draft-07/schema#"),
                Dialect.from_str("http://json-schema.org/draft-06/schema#"),
            ],
            description=(
                "A benchmark for validation of schemas containing "
                "lots of useless keywords. "
                "Checks we filter them out once, ahead of time."
            ),
            schema=dict(
                [
                    ("not", {"const": 42}),
                    *((str(i), i) for i in range(num_useless_keywords)),
                    ("type", "integer"),
                    *(
                        (str(i), i)
                        for i in range(
                            num_useless_keywords,
                            num_useless_keywords,
                        )
                    ),
                    ("minimum", 37),
                ],
            ),
            tests=[
                dict(description="Beginning of schema", instance=42),
                dict(description="Middle of schema", instance="foo"),
                dict(description="End of schema", instance=12),
                dict(description="Valid", instance=3737),
            ],
        ),
        uri=URL.parse(Path(__file__).absolute().as_uri()),
    )
